"""Segment actions
"""
import os
import sys

import wx

from atrip import Container, Collection
from atrip.compressor import find_compressor_by_name

from ..action import ViewerSubAction
from ..document import DiskImageDocument

from .. import commands
from .. import errors

import logging
log = logging.getLogger(__name__)


class compress_select(ViewerSubAction):
    def calc_enabled(self, action_key):
        return self.viewer.control.caret_handler.has_selection

    def perform(self, action_key):
        compressor_name = self.action_list_id
        compressor = find_compressor_by_name(compressor_name)
        log.debug(f"compressing with {compressor_name}")
        data = self.viewer.copy_data_from_selections()
        compressed = compressor.calc_packed_data(data)
        log.debug(f"data size={len(data)}, compressed={len(compressed)}")
        container = Container(compressed)
        collection = Collection(compressor_name, container=container)
        file_metadata = {'uri': compressor_name, 'mime': 'application/octet-stream', 'atrip_collection': collection}
        document = DiskImageDocument(file_metadata)
        self.viewer.frame.add_document(document)



import os

import numpy as np
import pytest

from atrip.memory_map import MemoryMap


def verify_map(memmap, test_data):
    for addr, label, count in test_data:
        item = memmap[addr]
        assert label == item[0]
        assert count == item[1]
        # print(f"tested {addr}, {label}")

class TestMemoryMap:
    def setup(self):
        pass

    @pytest.mark.parametrize(("filename", "rwtest", "rtest", "wtest"), [
            ["../omnivore/templates/atari800.labels", [(0x0012, "RTCLOK", 3), (0xd01a, "COLBK", 1), (0x200, "VDSLST", 2)], [(0xd000, "M0PF", 1), (0xd207, "POT7", 1)], [(0xd000, "HPOSP0", 1), (0xd207, "AUDC4", 1)]],
        ])
    def test_parse_from_file(self, filename, rwtest, rtest, wtest):
        rw, r, w = MemoryMap.from_file(filename)
        if rwtest is not None:
            verify_map(rw, rwtest)
        if rtest is not None:
            verify_map(r, rtest)
        if wtest is not None:
            verify_map(w, wtest)

    @pytest.mark.parametrize(("entries", "truth"), [
            [[(0, "LNFLG"), (1, "NGFLAG"), (2, "CASINI", "w"), (4, "RAMLO", "w"), (0x12, "RTCLOK", "3b")], [(0, "LNFLG", 1), (1, "NGFLAG", 1), (2, "CASINI", 2), (4, "RAMLO", 2), (0x0012, "RTCLOK", 3)]],
        ])
    def test_parse_from_list(self, entries, truth):
        labels = MemoryMap.from_list("Test", entries)
        verify_map(labels, truth)


if __name__ == "__main__":
    t = TestMemoryMap()
    t.setup()
    t.test_parse_from_file("../omnivore/templates/atari800.labels", [(0x0012, "RTCLOK", 3), (0xd01a, "COLBK", 1), (0x200, "VDSLST", 2)], [(0xd000, "M0PF", 1), (0xd207, "POT7", 1)], [(0xd000, "HPOSP0", 1), (0xd207, "AUDC4", 1)])



import sys

import numpy as np

import wx
import wx.adv
import wx.lib.buttons as buttons
import wx.lib.scrolledpanel
from wx.lib.expando import ExpandoTextCtrl
from wx.lib.stattext import GenStaticText  # standard static text can't set background color on some platforms

from atrip.char_mapping import internal_to_atascii, atascii_to_internal
from ..arch.ui.antic_colors import AnticColorDialog
from sawx.utils.runtime import get_all_subclasses

import logging
from functools import reduce
log = logging.getLogger(__name__)

def reduce_from_little_endian(raw):
    try:
        value = reduce(lambda x, y: x + (y << 8), raw)  #  convert to little endian
    except TypeError:
        # no array values, probably an attempt to display before the
        # segment is loaded.
        value = 0
    return value

class InfoField(object):
    keyword = ""
    same_line = False
    display_label = True
    use_edit_timer = False

    # wx.Sizer proportion of the main control (not the label).  See the
    # wx.Sizer docs, but basically 0 will fix vertical size to initial size, >
    # 0 will fill available space based on the total proportion in the sizer.
    vertical_proportion = 0

    default_width = 100

    popup_width = 300

    extra_vertical_spacing = 0

    edit_timer_delay = 300  # ms

    def __init__(self, panel, info):
        self.panel = panel
        self.set_args(info)
        self.create()

    @property
    def undo_label(self):
        return "Change %s" % self.field_name

    def set_args(self, args):
        self.field_name = args[0]
        self.byte_offset = args[1]
        self.byte_count = args[2]
        if len(args) > 3:
            val = args[3]
            try:
                self.max_val = int(val)
            except ValueError:
                self.attr_name_max_val = val

    def is_displayed(self, linked_base):
        return True

    def show(self, state=True):
        self.parent.Show(state)

    def hide(self):
        self.show(False)

    def enable(self, state=True):
        self.parent.Enable(state)
        self.ctrl.Enable(state)

    def create(self):
        self.parent = wx.Window(self.panel)
        self.box =  wx.BoxSizer(wx.VERTICAL)
        self.parent.SetSizer(self.box)
        if self.display_label:
            self.label = wx.StaticText(self.parent, label=self.field_name, style=wx.ST_ELLIPSIZE_END)
            bold_font = self.parent.GetFont()
            bold_font.SetWeight(weight=wx.FONTWEIGHT_BOLD)
            self.label.SetFont(bold_font)
        self.create_all_controls()
        if self.same_line and self.display_label:
            hbox = wx.BoxSizer(wx.HORIZONTAL)
            hbox.Add(self.label, 99, wx.ALIGN_CENTER)
            hbox.AddStretchSpacer(1)
            hbox.Add(self.ctrl, 0, wx.ALIGN_CENTER)
            for extra in self.extra_ctrls:
                hbox.Add(extra, 0, wx.ALIGN_CENTER)
            if self.extra_vertical_spacing > 0:
                self.box.AddSpacer(self.extra_vertical_spacing)
            self.box.Add(hbox, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
            if self.extra_vertical_spacing > 0:
                self.box.AddSpacer(self.extra_vertical_spacing)
        else:
            if self.display_label:
                self.box.Add(self.label, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
                self.box.AddSpacer(self.panel.LABEL_SPACING)
            self.box.Add(self.ctrl, self.vertical_proportion, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
            for extra in self.extra_ctrls:
                hbox.Add(extra, 0, wx.ALIGN_CENTER)
        self.box.AddSpacer(self.panel.VALUE_SPACING)

    def create_all_controls(self):
        if self.use_edit_timer:
            self.create_edit_timer()
        self.ctrl = self.create_control()
        if sys.platform.startswith("win"):
            self.ctrl.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel_scroll)
        self.extra_ctrls = self.create_extra_controls()
        self.set_control_limits()

    def create_edit_timer(self):
        self.edit_timer = wx.Timer(self.panel)
        self.panel.Bind(wx.EVT_TIMER, self.on_edit_timer_elapsed, self.edit_timer)

    def start_edit_timer(self, evt=None):
        self.edit_timer.Start(self.edit_timer_delay, oneShot=True)

    def on_edit_timer_elapsed(self, evt):
        self.edit_timer_callback()
        evt.Skip()

    def edit_timer_callback(self):
        raise RuntimeError("no edit timer callback set")

    def is_editable_control(self, ctrl):
        return ctrl == self.ctrl

    def set_control_limits(self):
        pass

    def create_extra_controls(self):
        return []

    def add_to_parent(self):
        self.panel.sizer.Add(self.parent, self.vertical_proportion, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.ALIGN_TOP, 0)
        self.show(True)

    def fill_data(self, linked_base):
        raise NotImplementedError

    def clear_data(self):
        raise NotImplementedError

    def has_focus(self):
        return self.ctrl.FindFocus() == self.ctrl

    def get_focus_params(self):
        pass

    def set_focus_params(self, params):
        pass

    def get_source_bytes(self, linked_base):
        raw = linked_base.segment[self.byte_offset:self.byte_offset+self.byte_count].copy()
        return raw

    def is_valid(self):
        return True

    def wants_focus(self):
        return False

    def set_focus(self):
        pass

    def on_mouse_wheel_scroll(self, event):
        screen_point = event.GetPosition()
        size = self.ctrl.GetSize()
        if screen_point.x < 0 or screen_point.y < 0 or screen_point.x > size.x or screen_point.y > size.y:
#            print "Mouse not over info panel %s: trying map!" % self
            self.panel.linked_base.control.on_mouse_wheel_scroll(event)
            return

        event.Skip()


class LabelField(InfoField):
    keyword = "label"
    same_line = True
    extra_vertical_spacing = 3

    def set_args(self, args):
        self.field_name = args[0]
        self.attr_name = args[1]
        self.max_val = args[2]

    def create_control(self):
        c = GenStaticText(self.parent, style=wx.ALIGN_RIGHT)
        return c

    def fill_data(self, linked_base):
        try:
            value = getattr(self.panel.linked_base.segment.jumpman_playfield_model, self.attr_name)
        except AttributeError:
            value = 0
        self.ctrl.SetLabel(str(value))
        self.set_background(linked_base, value <= self.max_val)

    def set_background(self, linked_base, valid):
        if valid:
            color = linked_base.cached_preferences.empty_background_color
        else:
            color = linked_base.cached_preferences.error_background_color
        self.ctrl.SetBackgroundColour(color)

    def clear_data(self):
        self.ctrl.SetLabel("")
        self.set_background(self.panel.linked_base, True)


class MultiLineLabelField(InfoField):
    keyword = "multi_line_label"
    same_line = False
    extra_vertical_spacing = 3

    def set_args(self, args):
        self.field_name = args[0]
        self.attr_name = args[1]

    def create_control(self):
        c = GenStaticText(self.parent, style=wx.ALIGN_LEFT)
        return c

    def fill_data(self, linked_base):
        try:
            value = getattr(self.panel.linked_base.segment.jumpman_playfield_model, self.attr_name)
        except AttributeError:
            value = ""
        self.ctrl.SetLabel(str(value).rstrip() + "\n")

    def clear_data(self):
        self.ctrl.SetLabel("")


class TextEditField(InfoField):
    keyword = "text"
    same_line = True
    wide = False
    use_edit_timer = True

    def create_control(self):
        if self.wide:
            size = (200, -1)
        else:
            size = (-1, -1)
        c = wx.TextCtrl(self.parent, size=size)
        c.Bind(wx.EVT_TEXT, self.start_edit_timer)
        c.SetEditable(True)
        return c

    def edit_timer_callback(self):
        linked_base = self.panel.linked_base
        self.process_text_change(linked_base)

    def set_control_limits(self):
        if self.byte_count > 0:
            self.ctrl.SetMaxLength(self.byte_count)

    def fill_data(self, linked_base):
        try:
            raw = self.get_source_bytes(linked_base)
            text = self.bytes_to_control_data(raw)
            self.ctrl.Enable(True)
        except IndexError:
            text = ""
            self.ctrl.Enable(False)
        except TypeError:
            text = ""
            self.ctrl.Enable(False)
        self.ctrl.ChangeValue(text)
        self.is_valid()

    def clear_data(self):
        self.ctrl.ChangeValue("")

    def get_focus_params(self):
        return self.ctrl.GetInsertionPoint()

    def set_focus_params(self, caret):
        self.ctrl.SetInsertionPoint(caret)

    def is_valid(self):
        c = self.ctrl
        c.SetBackgroundColour("#FFFFFF")
        try:
            self.parse_from_control()
            valid = True
        except Exception as e:
            log.debug("TextEditField: %s" % e)
            c.SetBackgroundColour("#FF8080")
            valid = False
        self.ctrl.Refresh()
        return valid

    def parse_from_control(self):
        return self.ctrl.GetValue()

    def initial_text_input(self, text):
        self.ctrl.ChangeValue(text)
        self.ctrl.SetInsertionPointEnd()#(self.ctrl.GetLastPosition())

    def bytes_to_control_data(self, raw):
        raw = self.map_bytes_to_text(raw)
        text = raw.tobytes().lstrip()
        if self.has_focus():
            data = self.parse_from_control()
            text = text[0:len(data)]
        else:
            text = text.rstrip()
        return text

    def map_bytes_to_text(self, raw):
        raw[raw < 32] = 32
        r = raw > 95
        raw[r] = raw[r] - 32  # convert lower case to upper
        raw[raw > 96] = 32
        return raw

    def parsed_to_bytes(self, parsed_data):
        raw = np.zeros([self.byte_count],dtype=np.uint8)
        text = np.fromstring(parsed_data, dtype=np.uint8)
        text = text[0:self.byte_count]
        text = self.map_parsed_to_bytes(text)
        raw[0:len(text)] = text
        return raw

    def map_parsed_to_bytes(self, text):
        return text

    def process_text_change(self, linked_base):
        if self.is_valid():
            data = self.parse_from_control()
            raw = self.parsed_to_bytes(data)
            linked_base.change_bytes(self.byte_offset, self.byte_offset + self.byte_count, raw, self.undo_label)


class AtasciiC0(TextEditField):
    keyword = "atascii_gr2_0xc0"
    high_bits = 0xc0
    wide = True

    def parsed_to_bytes(self, parsed_data):
        raw = np.zeros([self.byte_count],dtype=np.uint8) | self.high_bits
        text = np.fromstring(parsed_data, dtype=np.uint8)
        text = text[0:self.byte_count]
        text = self.map_parsed_to_bytes(text)
        # Center text
        i = (self.byte_count - len(text)) // 2
        raw[i:i + len(text)] = text
        return raw

    def map_bytes_to_text(self, raw):
        raw = raw & 0x3f
        return internal_to_atascii[raw]

    def map_parsed_to_bytes(self, text):
        text = atascii_to_internal[text]
        return text | self.high_bits


class UIntEditField(TextEditField):
    keyword = "uint"
    same_line = True

    def set_control_limits(self):
        pass

    def bytes_to_control_data(self, raw):
        value = reduce_from_little_endian(raw)
        text = str(value)
        return text

    def parse_from_control(self):
        value = int(self.ctrl.GetValue())
        if hasattr(self, "attr_name_max_val") and hasattr(self.panel.linked_base, self.attr_name_max_val):
            maxval = getattr(self.panel.linked_base.segment.jumpman_playfield_model, self.attr_name_max_val)
            if value > maxval:
                raise ValueError("%d out of range for attribute %s max of %d" % (value, self.attr_name_max_val, maxval))
        if hasattr(self, "max_val"):
            if value > self.max_val:
                raise ValueError("%d greater than %d" % (value, self.maxval))
        if self.byte_count == 1 and value >=0 and value < 256:
            return value
        elif self.byte_count == 2 and value >=0 and value < 256 * 256:
            return value
        raise ValueError("%d out of range for %d bytes" % (value, self.byte_count))

    def parsed_to_bytes(self, parsed_data):
        raw = np.empty([self.byte_count], dtype=np.uint8)
        if self.byte_count == 1:
            raw[0] = parsed_data
        elif self.byte_count == 2:
            v = raw.view(dtype='<u2')
            v[0] = parsed_data
        return raw


class AnticColorsField(InfoField):
    keyword = "antic_colors"
    same_line = True
    register_names = ["%x %s" % (i + 0x282a, n) for i, n in enumerate([
    "(unused)",
    "Jumpman Shadow",
    "Player 2",
    "Player 3",
    "Girder & up ropes",
    "Ladder & down ropes",
    "Coins",
    "Bullets",
    "Background",
    ])]

    def create_control(self):
        c = wx.Button(self.parent, -1, "Set Colors")
        c.Bind(wx.EVT_BUTTON, self.on_colors)
        return c

    def fill_data(self, linked_base):
        pass

    def clear_data(self):
        pass

    def on_colors(self, evt):
        linked_base = self.panel.linked_base
        raw = self.get_source_bytes(linked_base)
        dlg = AnticColorDialog(self.ctrl, raw, linked_base.cached_preferences, self.register_names)
        if dlg.ShowModal() == wx.ID_OK:
            linked_base.change_bytes(self.byte_offset, self.byte_offset + self.byte_count, dlg.colors, self.undo_label)
            # # linked_base.force_data_model_update()
            linked_base.document.byte_style_changed_event(flags=True)


class DropDownField(InfoField):
    keyword = "dropdown"
    same_line = True

    def set_args(self, args):
        self.field_name = args[0]
        self.byte_offset = args[1]
        self.byte_count = args[2]
        self.choices = args[3]

    def get_choices(self, linked_base):
        return self.choices

    def bytes_to_control_data(self, raw):
        value = reduce_from_little_endian(raw)
        value = min(value, len(self.choices) - 1)
        return value

    def fill_data(self, linked_base):
        choices = self.get_choices(linked_base)
        self.ctrl.SetItems(choices)
        raw = self.get_source_bytes(linked_base)
        default_choice = self.bytes_to_control_data(raw)
        self.ctrl.SetSelection(default_choice)

    def clear_data(self):
        pass

    def create_control(self):
        c = wx.Choice(self.parent, choices=[])
        c.Bind(wx.EVT_CHOICE, self.drop_down_changed)
        return c

    def drop_down_changed(self, event):
        raw = np.zeros([self.byte_count],dtype=np.uint8)
        raw[0] = self.ctrl.GetSelection()
        self.panel.linked_base.change_bytes(self.byte_offset, self.byte_offset + self.byte_count, raw, self.undo_label)


class CoinsNeededField(DropDownField):
    keyword = "coins_needed"
    same_line = True

    def bytes_to_control_data(self, raw):
        try:
            e = self.panel.linked_base.segment.jumpman_playfield_model
        except AttributeError:
            return 0
        num = e.num_coins
        value = reduce_from_little_endian(raw)
        diff = num - value
        if diff < 0:
            diff = 0
        diff = min(diff, len(self.choices) - 1)

        if e.coin_harvest_diff < 0:
            # First time! Calculate the offset to use subsequently
            e.coin_harvest_diff = diff
        return diff

    def drop_down_changed(self, event):
        e = self.panel.linked_base.segment.jumpman_playfield_model
        e.coin_harvest_diff = self.ctrl.GetSelection()
        raw = np.zeros([self.byte_count],dtype=np.uint8)
        raw[0] = max(0, e.num_coins - e.coin_harvest_diff)
        self.panel.linked_base.change_bytes(self.byte_offset, self.byte_offset + self.byte_count, raw, self.undo_label)


PANELTYPE = wx.lib.scrolledpanel.ScrolledPanel


class InfoPanel(PANELTYPE):

    """
    A panel for displaying and manipulating the properties of a linked_base.
    """
    LABEL_SPACING = 0
    VALUE_SPACING = 3
    SIDE_SPACING = 5

    def __init__(self, parent, linked_base, fields=None, **kwargs):
        self.linked_base = linked_base

        if fields is not None:
            self.fields = fields
        self.focus_on_input = None
        self.last_change_count = -1

        PANELTYPE.__init__(self, parent)

        # Mac/Win needs this, otherwise background color is black
        self.SetBackgroundColour(linked_base.cached_preferences.empty_background_color)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        self.init_fields(self.fields)

    def recalc_view(self):
        self.set_fields()

    def refresh_view(self):
        if self.IsShown():
            log.debug("refreshing %s" % self)
            self.recalc_view()
        else:
            log.debug("skipping refresh of hidden %s" % self)

    def init_fields(self, fields):
        self.sizer.AddSpacer(self.LABEL_SPACING)

        self.Freeze()

        focus = None
        self.current_fields = []
        for info in fields:
            field = self.create_field(info)
            if field is None:
                log.debug("init_fields: Skipping %s" % str(info))
                continue
            field.add_to_parent()
            field.fill_data(self.linked_base)
            if field.wants_focus():
                focus = field
            self.current_fields.append(field)

        self.constrain_size(focus)

        self.Thaw()
        self.Update()
        self.Refresh()

    def create_field(self, field_info):
        keyword = field_info[0]
        field = None
        subclasses = get_all_subclasses(InfoField)
        for kls in get_all_subclasses(InfoField):
            if keyword == kls.keyword:
                field = kls(self, field_info[1:])
                break
        return field

    def is_valid_data(self):
        """Override this in subclasses to disable entire panel if there is 
        no valid data to show.
        """
        return True

    def set_fields(self):
        enabled = self.is_valid_data()
        focus = None
        for field in self.current_fields:
            if field.has_focus():
                params = field.get_focus_params()
            else:
                params = None
            if enabled:
                field.fill_data(self.linked_base)
                if params is not None:
                    field.set_focus_params(params)
                    focus = field
            else:
                field.clear_data()
            field.enable(enabled)
        self.constrain_size(focus)

    def constrain_size(self, focus=None):
        self.sizer.Layout()
        self.focus_on_input = focus
        if focus is not None:
            self.ScrollChildIntoView(focus.ctrl)
        self.SetupScrolling(scroll_x=False, scrollToTop=False, scrollIntoView=True)

    def process_initial_key(self, event, text):
        """ Uses keyboard input from another control to set the focus to the
        previously noted info field and process the text there.
        """
        if self.focus_on_input is not None:
            self.focus_on_input.set_focus()
            self.ScrollChildIntoView(self.focus_on_input.ctrl)
            self.focus_on_input.initial_text_input(text)
            return True
        return False



import os
import glob

import numpy as np

from . import cputables
from .flags import *
from . import dtypes as ud
from . import libudis


class ParsedDisassembly(libudis.ParsedDisassembly):
    @property
    def entries(self):
        return self.raw_entries.view(dtype=ud.HISTORY_ENTRY_DTYPE)


class DisassemblyConfig(libudis.DisassemblyConfig):
    def get_parser(self, num_entries, origin, num_bytes):
        return ParsedDisassembly(num_entries, origin, num_bytes)


class Disassembler(object):
    def __init__(self, cpu_name, asm_syntax=None, memory_map=None, allow_undocumented=False, hex_lower=True, mnemonic_lower=False, r_mnemonics=None, w_mnemonics=None, rw_modes=None):
        self.source = None
        self.pc = 0
        self.pc_offset = 0
        self.origin = None
        self.memory_map = memory_map
        if asm_syntax is None:
            asm_syntax = {
                'comment char': ';',
                'origin': '.org',
                'data byte': '.db',
             }
        self.hex_lower = hex_lower
        self.mnemonic_lower = mnemonic_lower
        if mnemonic_lower:
            case_func = lambda a:a.lower()
        else:
            case_func = lambda a:a.upper()
        self.data_byte_opcode = case_func(asm_syntax['data byte'])
        self.asm_origin = case_func(asm_syntax['origin'])
        self.comment_char = case_func(asm_syntax['comment char'])
            
        self.setup(cpu_name, allow_undocumented, r_mnemonics, w_mnemonics, rw_modes)
    
    def setup(self, cpu_name, allow_undocumented, r_mnemonics, w_mnemonics, rw_modes):
        cpu = cputables.processors[cpu_name]
        modes = {}
        table = cpu['addressModeTable']
        if rw_modes is None:
            rw_modes = set()
        for mode, fmt in list(table.items()):
            if self.hex_lower:
                fmt = fmt.replace(":02X", ":02x").replace(":04X", ":04x")
            modes[mode] = fmt
        d = {}
        table = cpu['opcodeTable']
        for opcode, optable in list(table.items()):
            try:
                length, mnemonic, mode, flag = optable
            except ValueError:
                length, mnemonic, mode = optable
                flag = 0
            if mode in rw_modes:
                if mnemonic in r_mnemonics:
                    flag |= r
                if mnemonic in w_mnemonics:
                    flag |= w
            if not self.mnemonic_lower:
                mnemonic = mnemonic.upper()
            d[opcode] = (length, mnemonic, modes[mode], flag)
        self.ops = d
        self.leadin = cpu['leadInBytes']
        self.maxlen = cpu['maxLength']
        self.undocumented = allow_undocumented
    
    def get_data_byte_string(self, bytes):
        fmt = "$%02" + ("x" if self.hex_lower else "X")
        text = ",".join([fmt % b for b in bytes])
        return "%s %s" % (self.data_byte_opcode, text)
        
    def set_pc(self, source, pc):
        self.source = source
        self.length = len(source)
        self.pc = pc
        if self.origin is None:
            self.origin = pc
            self.pc_offset = -pc  # index into source array of pc
        
    def get_next(self):
        if self.pc >= self.origin + self.length:
            raise StopIteration
        opcode = int(self.source[self.pc + self.pc_offset])
        self.pc += 1
        return opcode
    
    def put_back(self):
        self.pc -= 1
    
    def get_style(self):
        return 0
    
    def is_data(self):
        """Check if the current PC location is a data byte"""
        return False
    
    def is_same_data_style(self, style, first_style, bytes):
        """Check if current PC is the same style of data as the first data byte
        in this block of data
        """
        return style == first_style
    
    def get_data_comment(self, style, bytes):
        comment = "%s $%x data byte" % (self.comment_char, len(bytes))
        if len(bytes) > 1:
            comment += "s"
        return comment
    
    def disasm_data(self):
        pc = self.pc
        # Read first byte to allow for StopIteration to signal end of file
        first_style = self.get_style()
        opcode = self.get_next()
        bytes = [opcode]
        try:
            # after first byte, read until bytes run out, end of data region,
            # or comment within data region
            while True:
                next_style = self.get_style()
                opcode = self.get_next()
                bytes.append(opcode)
                if not self.is_same_data_style(next_style, first_style, bytes):
                    self.put_back()
                    bytes.pop()
                    break
        except StopIteration:
            pass
        opstr = self.get_data_byte_string(bytes)
        comment = self.get_data_comment(first_style, bytes)
        return pc, bytes, opstr, comment, 0, None
    
    def disasm(self):
        pc = self.pc
        if self.is_data():
            return self.disasm_data()
        opcode = self.get_next()
        bytes = [opcode]
        
        try:
            # Handle if opcode is a leadin byte
            if opcode in self.leadin:
                next_pc = self.pc
                opcode2 = self.get_next()
                bytes.append(opcode2)
                opcode = (opcode << 8) + opcode2
                leadin = True
            else:
                leadin = False

            try:
                length, opstr, fmt, flag = self.ops[opcode]
            except KeyError:
                if len(bytes) == 2:
                    self.put_back()
                    opcode = bytes[0]
                    bytes = (opcode, )
                length, opstr, fmt, flag = 0, self.get_data_byte_string([opcode]), "", 0
            #print("0x%x" % opcode, fmt, length, opstr, mode, flag)
            if leadin:
                extra = length - 2
            else:
                extra = length - 1
        
            if flag & und and not self.undocumented:
                if len(bytes) == 2:
                    self.put_back()
                    opcode = bytes[0]
                    bytes = (opcode, )
                extra, opstr, fmt, flag = 0, self.get_data_byte_string([opcode]), "", 0
            
            next_pc = self.pc
            if extra == 1:
                operand1 = self.get_next()
                bytes.append(operand1)
                if flag & pcr:
                    signed = operand1 - 256 if operand1 > 127 else operand1
                    rel = (pc + 2 + signed) & 0xffff  # limit to 64k address space
                    opstr = opstr + " " + fmt.format(rel)
                    memloc = rel
                    dest_pc = rel
                else:
                    opstr = opstr + " " + fmt.format(operand1)
                    memloc = operand1
                    dest_pc = memloc
            elif extra == 2:
                operand1 = self.get_next()
                bytes.append(operand1)
                operand2 = self.get_next()
                bytes.append(operand2)
                if flag & z80bit:
                    opcode = (opcode << 16) + operand2
                    # reread opcode table for real format string
                    length, opstr, fmt, flag = self.ops[opcode]
                    signed = operand1 - 256 if operand1 > 127 else operand1
                    opstr = opstr + " " + fmt.format(signed)
                    memloc = None
                elif flag & pcr:
                    addr = operand1 + 256 * operand2
                    signed = addr - 0x10000 if addr > 32768 else addr
                    rel = (pc + 2 + signed) & 0xffff  # limit to 64k address space
                    memloc = rel
                    dest_pc = rel
                    try:
                        opstr = opstr + " " + fmt.format(rel)
                    except IndexError:
                        # maybe it's the 65c02 zeropagerelative where it's
                        # actually two separate operands
                        signed = operand2 - 256 if operand2 > 127 else operand2
                        rel = (pc + 2 + signed) & 0xffff  # limit to 64k address space
                        opstr = opstr + " " + fmt.format(operand1, rel)
                        memloc = rel
                        dest_pc = rel
                else:
                    opstr = opstr + " " + fmt.format(operand1, operand2)
                    memloc = operand1 + 256 * operand2
                dest_pc = memloc
            elif extra == 3:
                operand1 = self.get_next()
                bytes.append(operand1)
                operand2 = self.get_next()
                bytes.append(operand2)
                operand3 = self.get_next()
                bytes.append(operand3)
                opstr = opstr + " " + fmt.format(operand1, operand2, operand3)
                memloc = operand1 + 256 * operand2 + 256 * 256 * operand3
                dest_pc = memloc
            else:
                if fmt:
                    opstr = opstr + " " + fmt
                memloc = None
                dest_pc = None
            
        except StopIteration:
            self.pc = next_pc
            opstr, extra, flag = self.get_data_byte_string([opcode]), 0, 0
            memloc = None
            dest_pc = None
        if flag & r:
            rw = "r"
        elif flag & w:
            rw = "w"
        else:
            rw = ""
        
        bytes = tuple(bytes)
        comment = self.get_memloc_name(memloc, rw)
        is_und = self.flag_as_undefined(flag)
        return pc, bytes, opstr, comment, is_und, dest_pc
    
    get_instruction = disasm

    def get_disassembly(self):
        while True:
            result = self.disasm()
            print(f"disassembly result: {result}")
            yield result
    
    def get_memloc_name(self, memloc, rw):
        if rw == "" or self.memory_map is None:
            return ""
        flag = rw == "w"
        name = self.memory_map.get_name(memloc, flag)
        if name:
            return self.comment_char + " " + name
        return ""
    
    def flag_as_undefined(self, flag):
        return False


if __name__ == "__main__":
    import sys
    import argparse
    
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--cpu", help="Specify CPU type (defaults to 6502)", default="6502")
    parser.add_argument("-u", "--undocumented", help="Allow undocumented opcodes", action="store_true")
    parser.add_argument("-x", "--hex", help="Disassemble a string version of hex digits")
    parser.add_argument("filenames", metavar="filenames", nargs='*',
                   help="Binary files(s) to disassemble")
    args = parser.parse_args()

    def process(binary):
        pc = 0;
        disasm = Disassembler(args.cpu, allow_undocumented=args.undocumented)
        disasm.set_pc(binary, 0)
        for addr, bytes, opstr, comment, flag, dest_pc in disasm.get_disassembly():
            print("0x%04x %-12s ; %s   %s %s" % (addr, opstr, comment, bytes, flag))

    if args.hex:
        try:
            binary = args.hex.decode("hex")
        except TypeError:
            print("Invalid hex digits!")
            sys.exit()
        binary = np.fromstring(binary, dtype=np.uint8)
        process(binary)
    else:
        for filename in args.filenames:
            with open(filename, 'rb') as fh:
                binary = fh.read()
            binary = np.fromstring(binary, dtype=np.uint8)
            process(binary)



sha1_signatures = {
"application/x.atari5200.cart": {
  b'\xe1\x8a`;\x18\xb8^$\xf9\xc98\xb3;`w\xa7g\xc1\xfa\x9c': 'A.E. (USA) (Proto)',
  b'\xab10%m\x16\xaf\t\xca\xc8\xb1\x17I\xdf\xaf\xf6\xaa\xaa>\xa3': 'Blaster (USA) (Proto)',
  b'\xcf\xdd/\x8d\xab\xb0=\x87\x84\xf36\xbb\x0c\x80Bp\xdc:\x8al': 'Bounty Bob Strikes Back! (USA)',
  b'\xc1\xa5j\x1f\xc5\x1e\t\xb1\xd8\xe9:\xf4\r\xddu\x96\x84~ad': 'Castle Blast (USA) (Unl)',
  b'\x9f\x1e\xe1\x911\x9e\xf0&\x1c\xc7\x07J\x81\xc0\xe6\xd6\xb7FL\xd2': 'Choplifter! (USA)',
  b'\r\xf0\xd7\xf1\x1d\x89\xe2\xc5\xb5\x9c\xc9\xf9\x8dg\xd9\x8e 81\xaa': 'Galaxian (USA)',
  b'\x95G\xc5n)Q\xe1\xb3\xa7\x13\xb0\xb9\x9d>\xdbC\xc9\xd3\xcd\xd8': 'Koffi - Yellow Kopter (USA) (Unl)',
  b'W\xdf\xc2,\x98\x05U_\x95~\xd5\x1a-\xc7\xbf5\xad\x00\xc9\xd9': 'Star Raiders (USA)',
  b'j\xd7\xa1\xe8\xc9\xfa\xd4\x86\xfb\xec\x94\x98\xcbH\xbf[\xc3\xad\xc50': '[BIOS] Atari 5200 (USA)',
},
"application/x.atari5200.cart.type16": {
  b'\xdb 9s\xdf\x13\xb0q\xdaF\xd5/\xdb\x07\x8av\xf4\xdbB\x8d': 'Activision Decathlon, The (USA)',
  b'\xb7]U]\x05\x9a\xd6\x19\x92\xf9\xf5\xa24\x12\x9e\xcb\xf9PL5': 'Beamrider (USA)',
  b"\xe5l\xcck\xf4'\xa82\xce\x00T\x0b\x02\xd0e\xa2h\x82`\xdc": 'BerZerk (USA)',
  b'\x1e\x1a \xc1\x04\xbfMU\xec\xa0\xca\x1er\xdc]\xc6O\xa6\xb1\xb5': 'Blue Print (USA)',
  b'\xb0Q\xf0\xfb.\x81g\x87\xabn\xed\x0c\xb0\x08\xb9\xf8\xcc\x99c\xdc': "David Crane's Pitfall II - Lost Caverns (USA)",
  b'HbZ\xf4\xf3\xe9\xfeS\x8d\xab*\x0bF8\x87\xfc\x8ck\x13\x8d': 'Final Legacy (USA) (Proto)',
  b'\xd2\x86\xec\xb6y:\r\xda\x9c\x9ch\xf65]\xbe:\xdei"\xf9': 'H.E.R.O. (USA)',
  b'@F\xf6\x06\x02%J}\x86\xd2\xa6\xef\xd1\xa30E^\xae\xc1\x19': 'Last Starfighter, The (USA) (Proto)',
  b"2\xd8j8\x1f\xef\xf41I\xb6c6\xdc\xc4\xba%#'2(": 'Meteorites (USA)',
  b"\x12C\xa8\xec\xfc\x02\xa7J\xe4\xe1\xf4\xfc\xda8P'v\xf4\\\xb0": 'Millipede (USA) (Proto)',
  b'\x05d\xb1\x86z\x0bW\rf\xdf\xcb\xc1\x1a\xdc>Q\xa2\xc6\xf2\x8c': 'Miner 2049er Starring Bounty Bob (USA)',
  b'\x02\xdd[\x92x\xc0j\xe8\xcb\xb6\xda\xcc\tN#\x8f\xba\xcf\x08\xd2': 'Moon Patrol (USA)',
  b'\x17\xa4\x1c_\x9a2\xf8\xa7\x199\xcf\xaa\x0e\x96\x80{\x95\x1a\xed\xad': 'Quest for Quintana Roo (USA)',
  b'\x91u\xe6\xda\xcaX\x10*\xc8\xb9r\x9d{\x036\xcf\t\x97}\x12': 'Robotron 2084 (USA)',
  b'Z\xe2\x0ev}\x98\x19\xfe`\x1f;\x1e\xa6\xd4Pj\xe2\xf2\xce\xef': 'Space Shuttle - A Journey Into Space (USA)',
  b'#\x8c\xb4HU]\x8b\xb5\xc9j\xa05G\x18A\x1e\x87w\xc6\x13': 'Super Pac-Man (USA) (Proto)',
  b'\x85\xfc\xea\nMM\x1f"Cx\x01\xc9b)\xba\x96\xe0\xf6\xb7\x11': 'Tempest (USA) (Proto)',
  b'\x9d\xfb\xb5\xd8\xb7u\x04\xc6l\xe3\xacC\x82\xda\xf7\x98 9=\xd8': 'Track and Field (USA) (Proto)',
  b'I\xde\xbd\xf3\xc13\xfd\xe9\x90]m\x917\xda]\x8d\xff\x8db\x83': 'Wizard of Wor (USA)',
  b'\x181[\xec\x1eb\x10"\x90\xfe\xebw\xfb)\x0e\x12f\xcf[\xfc': 'Zone Ranger (USA)',
},
"application/x.atari5200.cart.type19": {
  b'\xe6O\xb1\xa3S\xc7+\x87\xc5H;\xb8\x8c\xb3R>\xa5\xd3{Y': '5200 Menu (USA) (Proto)',
  b'\x18i}\xd8\xe6@\xfa\x15x\xb3\xb4\xae\t\xafHL\x14\xfb\x84\x10': 'Asteroids (USA) (Proto)',
  b'\xd3\xff\x99\xe0\x80\x98\\\x9c\x8e%8\x9d\xfat\x83^\x8c\x9b\xe5\x08': "Atari PAM - Pete's Test (USA)",
  b'\xaf\x81\xc4\x00\xf1\xaf\x0e\xe1:$\xda^\x8bd$B\xa0\xfc\x94y': 'Atari PAM System Test (USA) (v1.02)',
  b'\xce\x9b\xd3\xb0\x84|^\x18|>\x8ff|\x8f^\xa7q\xc4\x99e': "Carol Shaw's River Raid (USA)",
  b'\xea{@\xce\x11\x067\x8e\x1d@\x9c\xabf4\xfb.H?\xf28': 'Dreadnaught Factor, The (USA)',
  b'\xd9\xa6\x95u\x7f\x04\x94\xae\xe9*P\xab\x8aIU\x87E\xe0\xce\xdf': 'Frogger (USA)',
  b'\x80\x92k\xb2p\xeaf\xcb\\\xac\xc3y\xde\x1f\xc1m\x8a=\x81K': 'Gorf (USA)',
  b'\xc7\xeb+JF\xa1\x97\xad\xfe\xdf3\xc2\xcc\xc1}\x84$\xd6\x81\xc4': 'K-Razy Shoot-Out (USA)',
  b'\x9f1\xc8Z\x13\xe9\x19\x13T8Y\x06\x01\xe3;\xef\x11\xa8\xdf\x8a': 'Keystone Kapers (USA)',
  b'\xf5Q$`\xc6\xb6\xd3\xa1\xba\xa0\xb2:\xdfj:7\xfe0\x8e\xb3': 'MegaMania (USA)',
  b'\\\x8e\xc2.\x13\x18\x10\xcc\xac\x08\x06l\xa9n\x99D\xcc\xfaN\xcb': 'Missile Command (USA)',
  b'.z\xe5r`bM\x1fW\x10\xd4E\xa0\x93o\x0c(\nfU': 'Mountain King (USA)',
  b'\x00.o\x8cu3\xb0?\x80A\xe7\xf1\x89e\x89\xc0\xab\x0e\x17\xa6': "Mr. Do!'s Castle (USA)",
  b'\xd1\xc7\xdc=\xe0\x97\xa4\xbf\xab\x82#Z\xce\xc6\x9a\x97\xe6\xe0_\x18': 'Pitfall! (USA)',
  b'\xe6\x96k\xde\\\xd4\x16p \xc9\xd2\x1b\xaba:\xf1d\x8bMu': 'Q-bert (USA)',
  b'\xcfA\x1d\xb7p!{Z\xf5\xfc\xfa2r\\\xc5\x87\xa8\xb9\xe7\x10': 'Space Invaders (USA)',
  b'\xbfw\x04\x8a\xbc6\xba,\xba^\xee\xac\x81(\x02\xd6\xa0\xeb\xa8\x91': 'Star Wars - Return of the Jedi - Death Star Battle (USA)',
  b'\x82\x05\t?_\xb1\xfd\xf0\xcc\xb7\x1df~\xa0\xa3\x84\x9f\xb0\xa0\xf2': 'Super Cobra (USA)',
  b'U\xdaV;C\x94P\x8c#\xd8f\xfb\x1eT2b\xfbESm': 'Zenji (USA)',
},
"application/x.atari5200.cart.type20": {
  b'R6\xbc\x13\xe4\xb0m.\x16\xd7\x0b\n\x0c\x98X\x8e\x17&\xda\x0c': 'Boogie (USA) (Demo)',
  b'\xa8\x0f\x87j\x89\xa8\xce\xac\xc7\x1d\xa5\xd9\xc0+Tip\xb1w0': 'Kaboom! (USA)',
  b'\xf5\xf9\xc7\xf9{\xf0\x9eVG\xe8\x08\xd8\x92\xb6V\xf3\xee\x1e\x19\xe0': 'Super Breakout (USA)',
  b'\xe5\x8e\x1f%y#iM\xbe\xb5\x86\xc8`\xf3\x85\xe7NF\x0f\x90': 'Yellow Submarine (USA) (Demo)',
},
"application/x.atari5200.cart.type4": {
  b'\x1d\xfd\x9c\xb5\xeb`\x85\x90\xc8N\x147b\x91\xf8\xc0/\xb1\xab\x93': 'Ballblazer (USA)',
  b'Q\x0c\xcb \xdf.\xcd\xbe\x7f\x83s\xdej\x9f\xc1\x14\x93\xe8\xc3\xf2': 'Barroom Baseball (USA) (Proto)',
  b'\x9b~\xcdDKb9D.\xf1\xf5L\xd8\xe1L)L\x06M\xa2': 'Black Belt (USA) (Proto)',
  b"N\x02.\x10\x1dsF\xa0\xa8a\x80\x187'@\xe73;J6": 'Castle Crisis (USA) (Unl)',
  b'\xf3\xc3+9X\rX\x97\x99\xdc\xea\xd9\x8b\xb2\x06\xfb\xf3\xdf\x05\x04': 'Gremlins (USA)',
  b'\xca}]\x08q\xbe\x19\xca"\xd0\x10\x19\xfdQ+X6\xe3\x13\n': 'Joust (USA) 32k',
  b'\xa2\x93\xb6\xb0`P\xe1\x10)\xd9q7\xf2\xb1\xb3`YVrs': 'Mario Bros. (USA)',
  b'\xff\x0f\x1d\x9e\xd0\xaeE\xe3\xc8\x1b%/=\n3^\x03\xb0\xaaK': 'Meebzork (USA) (Proto)',
  b'\x19\xffv"\xdb\xbc)\x15\xe3\xe9\x8d\x8d\x96c@\xe7\xe8\x93\xff4': 'Pengo (USA)',
  b'\xa9M\x06s\x9f\xa1\x94S\xfaXj\x83n\xaa\xca\xe1\x1f\xe99\x89': 'RealSports Baseball (USA)',
  b'\x9b!\xac\x7f\xec\x8dUe\x85\xc6\x97\x0211\x8e\xf1\xf5\x01\xb2\xb5': 'RealSports Basketball (USA)',
  b'\x9e\xdb\x06j7\xcavW\x97\x99R\xc9F\xab\x89\xbey&\xf4\xf8': 'RealSports Basketball (USA) (83-10-31) (Proto)',
  b'\x92S\xd7(l\xf8y\x99GK7\xdf$+\xa2;CX\xefJ': 'Rescue on Fractalus! (USA)',
  b'h\x813v\x1cF-he\xbfg>\xd2\x98\xb4\xd8(\x12\t@': 'Spitfire (USA) (Proto)',
  b'lc=\xe5\x0el\xc0\xc0\xc4\x00a\xcd$\xe6\xf7\x8fU*C\x8f': 'Vanguard (USA)',
  b'\xc3\xcf3!m\xfaHC\t\xb9#\xf2\x18)\xd0@Hk\xc4\x81': 'Xevious (USA) (Proto)',
  b'JLJ%\xc8\xed\x1b\n\xe7\x9c\xa2&\x84\xd7\xd3\x11\xa5d\xa7\xb5': 'Zaxxon (USA)',
},
"application/x.atari5200.cart.type6": {
  b'g\xe4\x05\xc3~U\xde\xc4\x07\x10\x18\xe8\x16j\xb4\xfa\xf9\xf2\x1d\xf5': 'Astro Chase (USA)',
  b'\\\xc3\xb4$_\xaf\x01\xc3\x0fw\xac \xb0\xb4\x94\x0fO\x8f\x83\xd8': 'Atari PAM Diagnostics (USA) (v2.0)',
  b'\xe0\x86\xe9\x86\x8a\xa9\x13\xe8\x7f\x94go\xd6\x00\xf0\x03\x89.s\xa1': 'Atari PAM Diagnostics (USA) (v2.3)',
  b'\xa39\xcd\xdf\xc2\x84w\xc2\x1cXx=-;\x0c\xf0\xf4bm\xdf': 'Battlezone (USA) (Proto)',
  b'\xbc|\x81\xf7X\xdd>"NB\xac\x8a\x16\xc1\xa1-i\xae\xd7"': 'Buck Rogers - Planet of Zoom (USA)',
  b'\xeb4;\x90\xd3\xdeini\xb1\xf09\xd4\xfb,j\xfb\r\x80\xc0': 'Centipede (USA)',
  b'\xe0 <!$\x12~\xba-g\x9b\xe5H\xfa%~\xcc[&\x07': 'Congo Bongo (USA)',
  b'\xfamw\xe6\xe9\xaf\xd0\xeaB\xfc\x9f\x98\x07\x93+\xf9\xe8\xfb\x17\x0e': 'Countermeasure (USA)',
  b'D\xf6\xdb\xf5\x01:p\xe6F\xd5\xc5#)\x81\xd5\x17\x06\x9f\xc4I': 'Defender (USA)',
  b'8zv\xbaY\x12%\x1e"\x85\xb7r \x04\xadu\xe9\xed\xe5#': 'Dig Dug (USA)',
  b'i\xdc\xc8\xf5\x9c\xa4I\xf1=\xc8\x96\xb4\xb2\x9bSGb\xd9pd': 'Frisky Tom (USA) (Proto)',
  b'=\x81U\x94\xcfgF\x94\xb0K\xb0\xff\x90\xbfc]}@/\xc2': 'Frogger II - Threeedeep! (USA)',
  b'\xd0\xd8e\x0e\xb0}\x16y\x00MI\x13`\xc2\xc0A\xe3|a\xcd': 'Gyruss (USA)',
  b'\xaevYC\xae\x8d\x90\x9bl\xa6\xff\x15`\xba\x02z\xde\xea\xa5\x15': 'James Bond 007 (USA)',
  b'\x0ft\x12^\xda\xc9\x0f\xfd\x12t\xa1(\xdf)p\xc294\x8f\xb5': 'Joust (USA)',
  b'\xa3\xe7`\xd0\x9eM\xc6\x052\x86\x96uv\xb5\xb7\xd4.\xe2\xfd8': 'Jr. Pac-Man (USA) (Proto)',
  b'\xfa\x02\xee\x04\xcf\xad\x90\xc6\xcc\xf9\xd9\xb9\x84\xe3\xb7\xb32\x7f\xec\x12': 'Jungle Hunt (USA)',
  b'\xc3\xdb\xee\xd7\xad\xc0\xe2\x82\xbfv3*\xb5\xa3 \x95P\xe2m\x02': 'Kangaroo (USA)',
  b'\xbcJSf\x16\xacEh\x9b:\xd7\x90\xf6\xb19N\xbb\xd8\xcc~': 'Looney Tunes Hotel (USA) (Proto)',
  b'/\x15*\x85z\x8aX\xd6\x02\xb3\xc7\x17a-\x97DB>\xe4\x95': 'Microgammon SB (USA) (Proto)',
  b'\xefM>\xe7\xb4\xda.\xf7\x9coM\r\x183\xf1\x0c\x1e\xb7r^': 'Miniature Golf (USA) (Proto)',
  b'~*\x05bU\x98\xc7\xe6a_\x96\x9a\xa6\x95\xe6C\x15\xcd\x98F': "Montezuma's Revenge featuring Panama Joe (USA)",
  b'\xf9=rDMv\x07Rs\xe5\x08UT#\x93e0\x87QH': 'Ms. Pac-Man (USA)',
  b'\xeeK<o\x08\xaa\x9f\xdf\xabS.\xec\x98\x13c\x06\xb6\x92?u': 'Pac-Man (USA)',
  b'\xd5\x04vLX\x94\xee\x99\xc0\xb5\xe2\x02L\xbe\x83>\xbb\xa3\xab\x17': 'Pole Position (USA)',
  b'\xe6\x15\xaa\xa0\xc1\xd5j\xc8;\x91\xf6\x08yp\xfap\x8dd3\xbf': 'Popeye (USA)',
  b'xl\xdfT\xd1b^\r>\xa8\x81E\x06\xfd\xee\xba^N\xceH': 'QIX (USA)',
  b'+C\xf7M\xb4\x832r\xd2\x04\x8d\xe8\xfa2\xb8W\xaf\x8e\t\x0f': 'RealSports Basketball (USA) (82-11-05) (Proto)',
  b'\xa0\xdb\xd6T\xed;\xa2X<\xa1u\xa15i$\xd9R\x98\x84\xdf': 'RealSports Football (USA)',
  b'\xc9\x11\xcc\x91O9#\xd5\xae\xb1\xa8\x97\xe6\xdcJU\x1a\x9aS\xda': 'RealSports Soccer (USA)',
  b'.ltr\xe3y&\xcd\xfd=\x85ZK\xd3z\x18\xfa\x83\xd8\xcf': 'RealSports Tennis (USA)',
  b"\nz\xae\xf5\xb0\xed\xe9'\xce\xa1>rI\x06 ~\x1d\xb4\x80h": 'Road Runner (USA) (Proto)',
  b'\x12~\xd1\xc3\x1c>\x85\xd1\x02\x12\xee\xe7\xb9\xe6\xd7\x1a\xef}\xd8\x91': 'Space Dungeon (USA)',
  b'_\x03T\xd8y-e#\x13\xe5\xa0\xfa]9^\t\x1c\x80\x1d\x97': 'Sport Goofy (USA) (Proto)',
  b'\x19\x80\x14T\xab\x86\xf2\xa5\xcfi\x1f\xcdWzq\xad+\x8d\xc7\xd5': 'Star Trek - Strategic Operations Simulator (USA)',
  b'\xabgo\xfe\x05K#\xbd\xd4<\xee$\x98\xaf\xef\xaa\xbb\xf95[': 'Star Wars - The Arcade Game (USA)',
  b'\t\xd1\x92\xa3Y\x8c\xda\xdf\xb6\x1e\xce\xb40\xb3\xd5\x89.@\x96/': 'Stargate (USA) (Proto)',
  b'\x8bU\x12\xd3\xd2\xa9R\xc9\x9b\x93w\x91\x9f\x0e\x06H\x11a\xd8\x1e': 'Xari Arena (USA) (Proto)',
},
}  # end sha1_signatures



#! python3

"""This is an APNG module, which can create apng file from pngs

Reference:
http://littlesvr.ca/apng/
http://wiki.mozilla.org/APNG_Specification
https://www.w3.org/TR/PNG/
"""

import struct
import binascii
import itertools
import io

__version__ = "0.2.1"

try:
	import PIL.Image
except ImportError:
	# Without Pillow, apng can only handle PNG images
	pass

try:
	import numpy as np

	isnumpy = lambda data: hasattr(data, "flat")
except ImportError:
	isnumpy = lambda data: False

PNG_SIGN = b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"

# http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html#C.Summary-of-standard-chunks
CHUNK_BEFORE_IDAT = {
	"cHRM", "gAMA", "iCCP", "sBIT", "sRGB", "bKGD", "hIST", "tRNS", "pHYs",
	"sPLT", "tIME", "PLTE"
}

def is_png(png):
	"""Test if ``png`` is a valid PNG file by checking the signature.
	
	:arg png: If ``png`` is a :any:`path-like object` or :any:`file-like object`
		object, read the content into bytes.
	:type png: path-like, file-like, or bytes
	:rtype: bool
	"""
	if isinstance(png, str) or hasattr(png, "__fspath__"):
		with open(png, "rb") as f:
			png_header = f.read(8)		
	elif hasattr(png, "read"):
		position = png.tell()
		png_header = png.read(8)
		png.seek(position)
	elif isinstance(png, bytes):
		png_header = png[:8]
	elif isnumpy(png):
		png_header = bytes(png.flat[:8])
	else:
		raise TypeError("Must be file, bytes, or str but get {}"
				.format(type(png)))
			
	return png_header == PNG_SIGN
			
def chunks_read(b):
	"""Parse PNG bytes into different chunks, yielding (type, data). 
	
	@type is a string of chunk type.
	@data is the bytes of the chunk. Including length, type, data, and crc.
	"""
	# skip signature
	i = 8
	# yield chunks
	while i < len(b):
		data_len, = struct.unpack("!I", b[i:i+4])
		type = b[i+4:i+8].decode("latin-1")
		yield type, b[i:i+data_len+12]
		i += data_len + 12

def chunks(png):
	"""Yield ``(chunk_type, chunk_raw_data)`` from ``png``.
	
	.. note::
	
		``chunk_raw_data`` includes chunk length, type, and CRC.
	
	:arg png: If ``png`` is a :any:`path-like object` or :any:`file-like object`
		object, read the content into bytes.
	:type png: path-like, file-like, or bytes
	:rtype: Generator[tuple(str, bytes)]
	"""
	if not is_png(png):
		# convert to png
		if isinstance(png, bytes):
			with io.BytesIO(png) as f:
				with io.BytesIO() as f2:
					PIL.Image.open(f).save(f2, "PNG", optimize=True)
					png = f2.getvalue()
		elif isnumpy(png):
			with io.BytesIO() as f2:
				im = PIL.Image.fromarray(png).save(f2, "PNG", optimize=True)
				png = np.fromstring(f2.getvalue(), dtype=np.uint8)
		else:
			with io.BytesIO() as f2:
				PIL.Image.open(png).save(f2, "PNG", optimize=True)
				png = f2.getvalue()
	
	if isinstance(png, str) or hasattr(png, "__fspath__"):
		# path like
		with open(png, "rb") as f:
			png = f.read()		
	elif hasattr(png, "read"):
		# file like
		png = png.read()
	elif isnumpy(png):
		# numpy array
		png = png.tostring()
		
	return chunks_read(png)
		
def make_chunk(type, data):
	"""Create a raw chunk by composing chunk's ``type`` and ``data``. It
	calculates chunk length and CRC for you.

	:arg str type: PNG chunk type.
	:arg bytes data: PNG chunk data, **excluding chunk length, type, and CRC**.
	:rtype: bytes
	"""
	out = struct.pack("!I", len(data))
	data = type.encode("latin-1") + data
	out += data + struct.pack("!I", binascii.crc32(data) & 0xffffffff)
	return out
	
class PNG:
	"""Represent PNG image. This class should only be initiated with
	classmethods."""
	def __init__(self):
		self.hdr = None
		self.end = None
		self.width = None
		self.height = None
		self.chunks = []
		
	def init(self):
		"""Extract some info from chunks"""
		for type, data in self.chunks:
			if type == "IHDR":
				self.hdr = data
			elif type == "IEND":
				self.end = data
				
		if self.hdr:
			# grab w, h info
			self.width, self.height = struct.unpack("!II", self.hdr[8:16])
			
	@classmethod
	def open(cls, png):
		"""Open a PNG file.
		
		:arg png: See :func:`chunks`.
		:rtype: :class:`PNG`
		"""
		o = cls()
		o.chunks = list(chunks(png))
		o.init()
		return o
		
	@classmethod
	def from_chunks(cls, chunks):
		"""Construct PNG from raw chunks.
		
		:arg chunks: A list of ``(chunk_type, chunk_raw_data)``. Also see
			:func:`chunks`.
		:type chunks: list[tuple(str, bytes)]
		"""
		o = cls()
		o.chunks = chunks
		o.init()
		return o
		
	def to_bytes(self):
		"""Convert entire image to bytes.
		
		:rtype: bytes
		"""
		chunks = [PNG_SIGN]
		chunks.extend(c[1] for c in self.chunks)
		return b"".join(chunks)
		
	def save(self, file):
		"""Save entire image to a file.

		:arg file: The destination.
		:type file: path-like or file-like
		"""
		if isinstance(file, str) or hasattr(file, "__fspath__"):
			with open(file, "wb") as f:
				f.write(self.to_bytes())
		else:
			file.write(self.to_bytes())
		
class FrameControl:
	"""A data class holding fcTL info."""
	def __init__(self, width=None, height=None, x_offset=0, y_offset=0, delay=100, delay_den=1000, depose_op=1, blend_op=0):
		"""Parameters are assigned as object members. See `https://wiki.mozilla.org/APNG_Specification <https://wiki.mozilla.org/APNG_Specification#.60fcTL.60:_The_Frame_Control_Chunk>`_ for the detail of fcTL.
		"""
		self.width = width
		self.height = height
		self.x_offset = x_offset
		self.y_offset = y_offset
		self.delay = delay
		self.delay_den = delay_den
		self.depose_op = depose_op
		self.blend_op = blend_op
		
	def to_bytes(self):
		"""Convert to bytes.
		
		:rtype: bytes
		"""
		return struct.pack("!IIIIHHbb", self.width, self.height, self.x_offset, self.y_offset, self.delay, self.delay_den, self.depose_op, self.blend_op)
		
	@classmethod
	def from_bytes(cls, b):
		"""Contruct fcTL info from bytes.
		
		:arg bytes b: The length of ``b`` must be *28*, excluding sequence
			number and CRC.
		"""
		return cls(*struct.unpack("!IIIIHHbb", b))

class APNG:
	"""Represent APNG image."""
	def __init__(self, num_plays=0):
		"""APNG is composed by multiple PNGs, which can be inserted with 
		:meth:`append`.
		
		:arg int num_plays: Number of times to loop. 0 = infinite.
			
		:var frames: Frames of APNG, a list of ``(png, control)`` tuple.
		:vartype frames: list[tuple(PNG, FrameControl)]
		:var int num_plays: same as ``num_plays``.
		"""
		self.frames = []
		self.num_plays = num_plays

	def __str__(self):
		return f"APNG: {len(self.frames)} frames, # plays={self.num_plays}"
		
	def append(self, png, **options):
		"""Read a PNG file and append one frame.
		
		:arg png: See :meth:`PNG.open`.
		:arg options: See :class:`FrameControl`.
		"""
		png = PNG.open(png)
		control = FrameControl(**options)
		if control.width is None:
			control.width = png.width
		if control.height is None:
			control.height = png.height
		self.frames.append((png, control))
		
	def to_bytes(self):
		"""Convert entire image to bytes.
		
		:rtype: bytes
		"""
		
		# grab the chunks we needs
		out = [PNG_SIGN]
		# FIXME: it's tricky to define "other_chunks". HoneyView stop the 
		# animation if it sees chunks other than fctl or idat, so we put other
		# chunks to the end of the file
		other_chunks = []
		seq = 0
		
		# for first frame
		png, control = self.frames[0]
		
		# header
		out.append(png.hdr)
		
		# acTL
		out.append(make_chunk("acTL", struct.pack("!II", len(self.frames), self.num_plays)))
		
		# fcTL
		if control:
			out.append(make_chunk("fcTL", struct.pack("!I", seq) + control.to_bytes()))
			seq += 1
		
		# and others...
		idat_chunks = []
		for type, data in png.chunks:
			if type in ("IHDR", "IEND"):
				continue
			if type == "IDAT":
				# put at last
				idat_chunks.append(data)
				continue
			out.append(data)
		out.extend(idat_chunks)
		
		# FIXME: we should do some optimization to frames...
		# for other frames
		for png, control in self.frames[1:]:
			# fcTL
			out.append(
				make_chunk("fcTL", struct.pack("!I", seq) + control.to_bytes())
			)
			seq += 1
			
			# and others...
			for type, data in png.chunks:
				if type in ("IHDR", "IEND") or type in CHUNK_BEFORE_IDAT:
					continue
				elif type == "IDAT":
					# convert IDAT to fdAT
					out.append(
						make_chunk("fdAT", struct.pack("!I", seq) + data[8:-4])
					)
					seq += 1
				else:
					other_chunks.append(data)
		
		# end
		out.extend(other_chunks)
		out.append(png.end)
		
		return b"".join(out)
		
	@classmethod
	def from_files(cls, files, **options):
		"""Create APNG from multiple files.
		
		This is same as::
		
			im = APNG()
			for file in files:
				im.append(file, **options)
				
		:arg list files: A list of file. See :meth:`PNG.open`.
		:arg options: Options for :class:`FrameControl`.
		:rtype: APNG
		"""
		o = cls()
		for file in files:
			o.append(file, **options)
		return o
		
	@classmethod
	def open(cls, file):
		"""Open an APNG file.
		
		:arg file: See :func:`chunks`.
		:rtype: APNG
		"""
		hdr = None
		head_chunks = []
		end = ("IEND", make_chunk("IEND", b""))
		
		frame_chunks = []
		frames = []
		num_plays = 0
		
		control = None
		
		for type, data in chunks(file):
			if type == "IHDR":
				hdr = data
				frame_chunks.append((type, data))
			elif type == "acTL":
				num_frames, num_plays = struct.unpack("!II", data[8:-4])
				continue
			elif type == "fcTL":
				if any(type == "IDAT" for type, data in frame_chunks):
					# IDAT inside chunk, go to next frame
					frame_chunks.append(end)
					frames.append((PNG.from_chunks(frame_chunks), control))
					
					control = FrameControl.from_bytes(data[12:-4])
					hdr = make_chunk("IHDR", struct.pack("!II", control.width, control.height) + hdr[16:-4])
					frame_chunks = [("IHDR", hdr)]
				else:
					control = FrameControl.from_bytes(data[12:-4])
			elif type == "IDAT":
				frame_chunks.extend(head_chunks)
				frame_chunks.append((type, data))
			elif type == "fdAT":
				# convert to IDAT
				frame_chunks.extend(head_chunks)
				frame_chunks.append(("IDAT", make_chunk("IDAT", data[12:-4])))
			elif type == "IEND":
				# end
				frame_chunks.append(end)
				frames.append((PNG.from_chunks(frame_chunks), control))
				break
			elif type in CHUNK_BEFORE_IDAT:
				head_chunks.append((type, data))
			else:
				frame_chunks.append((type, data))
				
		o = cls()
		o.frames = frames
		o.num_plays = num_plays
		return o
		
	def save(self, file):
		"""Save entire image to a file.

		:arg file: The destination.
		:type file: path-like or file-like
		"""
		if isinstance(file, str) or hasattr(file, "__fspath__"):
			with open(file, "wb") as f:
				f.write(self.to_bytes())
		else:
			file.write(self.to_bytes())



import glob

import numpy as np

from mock import *

from atrip.container import guess_container
from atrip.media_type import Media, guess_media_type
from atrip import errors, find_container, find_collection

class TestContainer:
    container = None

    def setup(self):
        if self.container is None:
            filename = "dos_sd_test1.atr"
            pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
            self.__class__.container = find_container(pathname)

    def test_find_file(self):
        c = self.container
        dirent = c.find_dirent("A128.DAT", True)
        print(dirent)
        assert dirent is not None

        dirent = c.find_dirent("a1024.dat", False)
        print(dirent)
        assert dirent is not None

        dirent = c.find_dirent("a1024.dat", True)
        print(dirent)
        assert dirent is None

    def test_uuid(self):
        c = self.container
        segments = list(c.iter_segments())
        for s in segments:
            print(s, s.uuid)
            found = c.find_uuid(s.uuid)
            assert s == found


class TestCollection:
    collection = None

    def setup(self):
        if self.collection is None:
            filename = "dos_sd_test_collection.zip"
            pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
            self.__class__.collection = find_collection(pathname)

    @pytest.mark.parametrize("filename", [
            "A128.DAT",
            "D1:A128.DAT",
            "A1024.DAT",
            "D1:A1024.DAT",
            "D2:E256.DAT",
            "D2:B512.DAT",
            "D3:A8000.DAT",
            "D4:A15000.DAT",
        ])
    def test_find_file(self, filename):
        c = self.collection
        dirent = c.find_dirent(filename, True)
        assert dirent is not None

        lower = filename.lower()
        dirent = c.find_dirent(lower, False)
        assert dirent is not None

        dirent = c.find_dirent(lower, True)
        assert dirent is None

    @pytest.mark.parametrize("filename", [
            "sir_not_appearing_in_this_film",
            "D1:sir_not_appearing_in_this_film",
            "D3:B256.DAT",
            "D4:F4096.DAT",  # deleted file
        ])
    def test_missing_file(self, filename):
        c = self.collection
        dirent = c.find_dirent(filename, True)
        assert dirent is None

    def test_uuid(self):
        c = self.collection
        segments = []
        for item in c.containers:
            segments.extend(item.iter_segments())
        print(c.verbose_info)
        for s in segments:
            print(s, s.uuid)
            found = c.find_uuid(s.uuid)
            assert s == found


if __name__ == "__main__":
    t = TestContainer()
    t.setup()
    # t.test_find_file()
    # t.test_uuid()
    t = TestCollection()
    t.setup()
    t.test_find_file("D1:A128.DAT")
    # t.test_uuid()



import numpy as np

from .. import errors
from ..segment import Segment
from ..filesystem import VTOC, Dirent, Directory, Filesystem
from ..file_type import guess_file_type

try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False

import logging
log = logging.getLogger(__name__)


class AtariCasDirent(Dirent):
    extra_serializable_attributes = ['start_index:int', 'bytes_on_tape:int', 'num_chunks:int', 'baud:int', '_filename']

    def __init__(self, directory, file_num, start_index):
        self.parse_chunks(directory, start_index)
        Dirent.__init__(self, directory, file_num, start_index, self.bytes_on_tape, parent=directory.filesystem.media)
        self.parse_chunks(directory, start_index)

    def init_empty(self):
        super().init_empty()
        self.start_index = 0
        self.bytes_on_tape = 0
        self.num_chunks = 0
        self.baud = 600
        self._filename = b''

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.start_index == other.start_index and self.bytes_on_tape == other.bytes_on_tape and self.num_chunks == other.num_chunks

    @property
    def filename(self):
        return self._filename

    @property
    def catalog_entry(self):
        return "%d %s  %03d" % (self.start_index, self.filename, self.num_chunks)

    @property
    def status(self):
        return f"{self.baud} baud"

    @property
    def verbose_info(self):
        return self.status

    def parse_chunks(self, parent, index):
        self.start_index = index
        media = parent.filesystem.media
        chunk = media.get_chunk(index)
        log.debug(f"Reading chunk {chunk}")
        if chunk.chunk_type != "FUJI":
            raise errors.FileStructureError(f"Expecting FUJI chunk to begin file")
        self._filename = chunk.chunk_data.copy()
        index += chunk.record_length
        offsets = np.empty(parent.filesystem.max_file_size, dtype=np.uint32)
        offset_index = 0
        keep_reading = True
        while keep_reading:
            try:
                chunk = media.get_chunk(index)
            except errors.InvalidSectorNumber:
                # EOF; maybe it's a tape without the 0xfe record?
                break
            log.debug(f"Reading chunk {chunk}")
            if chunk.chunk_type == "FUJI":
                # somehow encountered another start chunk; previous was missing
                # an EOF record
                log.warning("Missing EOF record; assuming end of file")
                break
            elif chunk.chunk_type == "baud":
                self.baud = chunk.chunk_aux
            elif chunk.chunk_type == "data":
                d = chunk.chunk_data
                flag = d[2]
                data_size = chunk.chunk_length - 8 - 3 - 1
                if not (flag == 0xfa or flag == 0xfc or flag == 0xfe):
                    raise errors.FileStructureError(f"Unknown data record flag {flag}")
                if flag == 0xfe:
                    # end of file record
                    keep_reading = False
                offsets[offset_index:offset_index + data_size] = np.arange(index, index + data_size)
                offset_index += data_size
            index += chunk.record_length
        self.temporary_offset_storage = np.copy(offsets[0:offset_index])
        self.bytes_on_tape = index - self.start_index

    def parse_raw_dirent(self):
        pass

    def get_file(self):
        media = self.filesystem.media
        offsets = self.temporary_offset_storage
        del self.temporary_offset_storage
        if len(offsets) > 0:
            file_segment = guess_file_type(media, self.filename, offsets)
            self.segments = [file_segment]
            return file_segment


class AtariCasDirectory(Directory):
    def find_segment_location(self):
        return 0, 0

    def calc_dirents(self):
        segments = []
        start_index = 0
        filenum = 0
        try:
            while True:
                dirent = AtariCasDirent(self, filenum, start_index)
                segments.append(dirent)
                start_index = dirent.start_index + dirent.bytes_on_tape
                filenum += 1
        except (errors.MediaError, errors.FileError):
            pass
        return segments


class AtariCassetteFilesystem(Filesystem):
    ui_name = "Atari Cassette (.cas)"
    default_executable_extension = "XEX"

    def check_media(self, media):
        try:
            media.get_chunk
        except AttributeError:
            raise errors.IncompatibleMediaError(f"{self.ui_name} only valid on cassette media")

    def calc_directory_segment(self):
        return AtariCasDirectory(self)



import textwrap

import numpy as np

from .. import errors
from ..stringifier import Stringifier


class BasicData(Stringifier):
    text_type = "basic_data"
    ui_name = "BASIC DATA statements"

    def calc_text(self, byte_data):
        lines = []
        start_line = "DATA "
        line_length = 38 - len(start_line)
        values = " ".join([str(int(i)) for i in byte_data])
        lines = textwrap.wrap(values, line_length)
        lines = [start_line + line.replace(" ", ",") for line in lines]
        text = "\n".join(lines)
        return text



START_FLAG = 0x1
SELECT_FLAG = 0x2
OPTION_FLAG = 0x4
SHIFT_FLAG = 0x10
CTRL_FLAG = 0x20
MOUSE_DELTA_FLAG = 0x80  # if not set, direct mouse mode is used where values are sent directly to POKEY potentiometer input; otherwise mouse movements are sent as deltas

# Atari key codes from akey.h

# All key codes are identical to the values in akey.h except for the
# special keys. In akey.h they are negative, but here they are positive
# because we're using unsigned bytes for all the code

AKEY_NONE = -1

# Special key codes.
AKEY_WARMSTART = 2
AKEY_COLDSTART = 3
AKEY_EXIT = 4
AKEY_BREAK = 5
AKEY_UI = 7
AKEY_SCREENSHOT = 8
AKEY_SCREENSHOT_INTERLACE = 9
AKEY_START = 10
AKEY_SELECT = 11
AKEY_OPTION = 12
AKEY_PBI_BB_MENU = 13
AKEY_CX85_1 = 14
AKEY_CX85_2 = 15
AKEY_CX85_3 = 16
AKEY_CX85_4 = 17
AKEY_CX85_5 = 18
AKEY_CX85_6 = 19
AKEY_CX85_7 = 20
AKEY_CX85_8 = 21
AKEY_CX85_9 = 22
AKEY_CX85_0 = 23
AKEY_CX85_PERIOD = 24
AKEY_CX85_MINUS = 25
AKEY_CX85_PLUS_ENTER = 26
AKEY_CX85_ESCAPE = 27
AKEY_CX85_NO = 28
AKEY_CX85_DELETE = 29
AKEY_CX85_YES = 30
AKEY_TURBO = 31

AKEY_SHFT = 0x40
AKEY_CTRL = 0x80
AKEY_SHFTCTRL = 0xc0

AKEY_0 = 0x32
AKEY_1 = 0x1f
AKEY_2 = 0x1e
AKEY_3 = 0x1a
AKEY_4 = 0x18
AKEY_5 = 0x1d
AKEY_6 = 0x1b
AKEY_7 = 0x33
AKEY_8 = 0x35
AKEY_9 = 0x30

AKEY_CTRL_0 = (AKEY_CTRL | AKEY_0)
AKEY_CTRL_1 = (AKEY_CTRL | AKEY_1)
AKEY_CTRL_2 = (AKEY_CTRL | AKEY_2)
AKEY_CTRL_3 = (AKEY_CTRL | AKEY_3)
AKEY_CTRL_4 = (AKEY_CTRL | AKEY_4)
AKEY_CTRL_5 = (AKEY_CTRL | AKEY_5)
AKEY_CTRL_6 = (AKEY_CTRL | AKEY_6)
AKEY_CTRL_7 = (AKEY_CTRL | AKEY_7)
AKEY_CTRL_8 = (AKEY_CTRL | AKEY_8)
AKEY_CTRL_9 = (AKEY_CTRL | AKEY_9)

AKEY_a = 0x3f
AKEY_b = 0x15
AKEY_c = 0x12
AKEY_d = 0x3a
AKEY_e = 0x2a
AKEY_f = 0x38
AKEY_g = 0x3d
AKEY_h = 0x39
AKEY_i = 0x0d
AKEY_j = 0x01
AKEY_k = 0x05
AKEY_l = 0x00
AKEY_m = 0x25
AKEY_n = 0x23
AKEY_o = 0x08
AKEY_p = 0x0a
AKEY_q = 0x2f
AKEY_r = 0x28
AKEY_s = 0x3e
AKEY_t = 0x2d
AKEY_u = 0x0b
AKEY_v = 0x10
AKEY_w = 0x2e
AKEY_x = 0x16
AKEY_y = 0x2b
AKEY_z = 0x17

AKEY_A = (AKEY_SHFT | AKEY_a)
AKEY_B = (AKEY_SHFT | AKEY_b)
AKEY_C = (AKEY_SHFT | AKEY_c)
AKEY_D = (AKEY_SHFT | AKEY_d)
AKEY_E = (AKEY_SHFT | AKEY_e)
AKEY_F = (AKEY_SHFT | AKEY_f)
AKEY_G = (AKEY_SHFT | AKEY_g)
AKEY_H = (AKEY_SHFT | AKEY_h)
AKEY_I = (AKEY_SHFT | AKEY_i)
AKEY_J = (AKEY_SHFT | AKEY_j)
AKEY_K = (AKEY_SHFT | AKEY_k)
AKEY_L = (AKEY_SHFT | AKEY_l)
AKEY_M = (AKEY_SHFT | AKEY_m)
AKEY_N = (AKEY_SHFT | AKEY_n)
AKEY_O = (AKEY_SHFT | AKEY_o)
AKEY_P = (AKEY_SHFT | AKEY_p)
AKEY_Q = (AKEY_SHFT | AKEY_q)
AKEY_R = (AKEY_SHFT | AKEY_r)
AKEY_S = (AKEY_SHFT | AKEY_s)
AKEY_T = (AKEY_SHFT | AKEY_t)
AKEY_U = (AKEY_SHFT | AKEY_u)
AKEY_V = (AKEY_SHFT | AKEY_v)
AKEY_W = (AKEY_SHFT | AKEY_w)
AKEY_X = (AKEY_SHFT | AKEY_x)
AKEY_Y = (AKEY_SHFT | AKEY_y)
AKEY_Z = (AKEY_SHFT | AKEY_z)

AKEY_CTRL_a = (AKEY_CTRL | AKEY_a)
AKEY_CTRL_b = (AKEY_CTRL | AKEY_b)
AKEY_CTRL_c = (AKEY_CTRL | AKEY_c)
AKEY_CTRL_d = (AKEY_CTRL | AKEY_d)
AKEY_CTRL_e = (AKEY_CTRL | AKEY_e)
AKEY_CTRL_f = (AKEY_CTRL | AKEY_f)
AKEY_CTRL_g = (AKEY_CTRL | AKEY_g)
AKEY_CTRL_h = (AKEY_CTRL | AKEY_h)
AKEY_CTRL_i = (AKEY_CTRL | AKEY_i)
AKEY_CTRL_j = (AKEY_CTRL | AKEY_j)
AKEY_CTRL_k = (AKEY_CTRL | AKEY_k)
AKEY_CTRL_l = (AKEY_CTRL | AKEY_l)
AKEY_CTRL_m = (AKEY_CTRL | AKEY_m)
AKEY_CTRL_n = (AKEY_CTRL | AKEY_n)
AKEY_CTRL_o = (AKEY_CTRL | AKEY_o)
AKEY_CTRL_p = (AKEY_CTRL | AKEY_p)
AKEY_CTRL_q = (AKEY_CTRL | AKEY_q)
AKEY_CTRL_r = (AKEY_CTRL | AKEY_r)
AKEY_CTRL_s = (AKEY_CTRL | AKEY_s)
AKEY_CTRL_t = (AKEY_CTRL | AKEY_t)
AKEY_CTRL_u = (AKEY_CTRL | AKEY_u)
AKEY_CTRL_v = (AKEY_CTRL | AKEY_v)
AKEY_CTRL_w = (AKEY_CTRL | AKEY_w)
AKEY_CTRL_x = (AKEY_CTRL | AKEY_x)
AKEY_CTRL_y = (AKEY_CTRL | AKEY_y)
AKEY_CTRL_z = (AKEY_CTRL | AKEY_z)

AKEY_CTRL_A = (AKEY_CTRL | AKEY_A)
AKEY_CTRL_B = (AKEY_CTRL | AKEY_B)
AKEY_CTRL_C = (AKEY_CTRL | AKEY_C)
AKEY_CTRL_D = (AKEY_CTRL | AKEY_D)
AKEY_CTRL_E = (AKEY_CTRL | AKEY_E)
AKEY_CTRL_F = (AKEY_CTRL | AKEY_F)
AKEY_CTRL_G = (AKEY_CTRL | AKEY_G)
AKEY_CTRL_H = (AKEY_CTRL | AKEY_H)
AKEY_CTRL_I = (AKEY_CTRL | AKEY_I)
AKEY_CTRL_J = (AKEY_CTRL | AKEY_J)
AKEY_CTRL_K = (AKEY_CTRL | AKEY_K)
AKEY_CTRL_L = (AKEY_CTRL | AKEY_L)
AKEY_CTRL_M = (AKEY_CTRL | AKEY_M)
AKEY_CTRL_N = (AKEY_CTRL | AKEY_N)
AKEY_CTRL_O = (AKEY_CTRL | AKEY_O)
AKEY_CTRL_P = (AKEY_CTRL | AKEY_P)
AKEY_CTRL_Q = (AKEY_CTRL | AKEY_Q)
AKEY_CTRL_R = (AKEY_CTRL | AKEY_R)
AKEY_CTRL_S = (AKEY_CTRL | AKEY_S)
AKEY_CTRL_T = (AKEY_CTRL | AKEY_T)
AKEY_CTRL_U = (AKEY_CTRL | AKEY_U)
AKEY_CTRL_V = (AKEY_CTRL | AKEY_V)
AKEY_CTRL_W = (AKEY_CTRL | AKEY_W)
AKEY_CTRL_X = (AKEY_CTRL | AKEY_X)
AKEY_CTRL_Y = (AKEY_CTRL | AKEY_Y)
AKEY_CTRL_Z = (AKEY_CTRL | AKEY_Z)

AKEY_HELP = 0x11
AKEY_DOWN = 0x8f
AKEY_LEFT = 0x86
AKEY_RIGHT = 0x87
AKEY_UP = 0x8e
AKEY_BACKSPACE = 0x34
AKEY_DELETE_CHAR = 0xb4
AKEY_DELETE_LINE = 0x74
AKEY_INSERT_CHAR = 0xb7
AKEY_INSERT_LINE = 0x77
AKEY_ESCAPE = 0x1c
AKEY_ATARI = 0x27
AKEY_CAPSLOCK = 0x7c
AKEY_CAPSTOGGLE = 0x3c
AKEY_TAB = 0x2c
AKEY_SETTAB = 0x6c
AKEY_CLRTAB = 0xac
AKEY_RETURN = 0x0c
AKEY_SPACE = 0x21
AKEY_EXCLAMATION = 0x5f
AKEY_DBLQUOTE = 0x5e
AKEY_HASH = 0x5a
AKEY_DOLLAR = 0x58
AKEY_PERCENT = 0x5d
AKEY_AMPERSAND = 0x5b
AKEY_QUOTE = 0x73
AKEY_AT = 0x75
AKEY_PARENLEFT = 0x70
AKEY_PARENRIGHT = 0x72
AKEY_LESS = 0x36
AKEY_GREATER = 0x37
AKEY_EQUAL = 0x0f
AKEY_QUESTION = 0x66
AKEY_MINUS = 0x0e
AKEY_PLUS = 0x06
AKEY_ASTERISK = 0x07
AKEY_SLASH = 0x26
AKEY_COLON = 0x42
AKEY_SEMICOLON = 0x02
AKEY_COMMA = 0x20
AKEY_FULLSTOP = 0x22
AKEY_UNDERSCORE = 0x4e
AKEY_BRACKETLEFT = 0x60
AKEY_BRACKETRIGHT = 0x62
AKEY_CIRCUMFLEX = 0x47
AKEY_BACKSLASH = 0x46
AKEY_BAR = 0x4f
AKEY_CLEAR = (AKEY_SHFT | AKEY_LESS)
AKEY_CARET = (AKEY_SHFT | AKEY_ASTERISK)
AKEY_F1 = 0x03
AKEY_F2 = 0x04
AKEY_F3 = 0x13
AKEY_F4 = 0x14

# Following keys cannot be read with both shift and control pressed:
#   J K L ; + * Z X C V B F1 F2 F3 F4 HELP

# 5200 key codes
AKEY_5200_START = 0x39
AKEY_5200_PAUSE = 0x31
AKEY_5200_RESET = 0x29
AKEY_5200_0 = 0x25
AKEY_5200_1 = 0x3f
AKEY_5200_2 = 0x3d
AKEY_5200_3 = 0x3b
AKEY_5200_4 = 0x37
AKEY_5200_5 = 0x35
AKEY_5200_6 = 0x33
AKEY_5200_7 = 0x2f
AKEY_5200_8 = 0x2d
AKEY_5200_9 = 0x2b
AKEY_5200_HASH = 0x23
AKEY_5200_ASTERISK = 0x27



from collections import defaultdict

import numpy as np

from . import errors
from .segments import SegmentData, EmptySegment, ObjSegment
from .diskimages import DiskImageBase
from .utils import to_numpy

import logging
log = logging.getLogger(__name__)


# From atari800 source
known_cart_types = [
# (note: all size units in KB)
# atari800 index number
# name
# total size
# static size
# static offset
# static address
# banked size
# banked offset (for bank zero)
# banked address
    (0,  "", 0,),
    (57, "Standard 2 KB", 2, 2, 0, 0xb800),
    (58, "Standard 4 KB", 4, 4, 0, 0xb000),
    (59, "Right slot 4 KB", 4, 4, 0, 0, 0x9000),
    (1,  "Standard 8 KB", 8, 8, 0, 0xa000),
    (21, "Right slot 8 KB", 8,),
    (2,  "Standard 16 KB", 16, 16, 0, 0x8000),
    (44, "OSS 8 KB", 8,),
    (15, "OSS one chip 16 KB", 16,),
    (3,  "OSS two chip (034M) 16 KB", 16, 4, 12, 0xb000, 4, 0, 0xa000),
    (45, "OSS two chip (043M) 16 KB", 16, 4, 12, 0xb000, 4, 0, 0xa000),
    (12, "XEGS 32 KB", 32,  8, 24, 0xa000, 8, 0, 0x8000),
    (13, "XEGS (banks 0-7) 64 KB", 64, 8, 56, 0xa000, 8, 0, 0x8000),
    (67, "XEGS (banks 8-15) 64 KB", 64, 8, 56, 0xa000, 8, 0, 0x8000),
    (14, "XEGS 128 KB", 128, 8, 120, 0xa000, 8, 0, 0x8000),
    (23, "XEGS 256 KB", 256, 8, 248, 0xa000, 8, 0, 0x8000),
    (24, "XEGS 512 KB", 512, 8, 504, 0xa000, 8, 0, 0x8000),
    (25, "XEGS 1 MB", 1024, 8, 1016, 0xa000, 8, 0, 0x8000 ),
    (33, "Switchable XEGS 32 KB", 32,  8, 24, 0xa000, 8, 0, 0x8000),
    (34, "Switchable XEGS 64 KB", 64,  8, 56, 0xa000, 8, 0, 0x8000),
    (35, "Switchable XEGS 128 KB", 128, 8, 120, 0xa000, 8, 0, 0x8000),
    (36, "Switchable XEGS 256 KB", 256, 8, 248, 0xa000, 8, 0, 0x8000),
    (37, "Switchable XEGS 512 KB", 512, 8, 504, 0xa000, 8, 0, 0x8000),
    (38, "Switchable XEGS 1 MB", 1024, 8, 1016, 0xa000, 8, 0, 0x8000 ),
    (22, "Williams 32 KB", 32,),
    (8,  "Williams 64 KB", 64,),
    (9,  "Express 64 KB", 64,),
    (10, "Diamond 64 KB", 64,),
    (11, "SpartaDOS X 64 KB", 64,),
    (43, "SpartaDOS X 128 KB", 128,),
    (17, "Atrax 128 KB", 128,),
    (18, "Bounty Bob 40 KB", 40,),
    (26, "MegaCart 16 KB", 16,),
    (27, "MegaCart 32 KB", 32,),
    (28, "MegaCart 64 KB", 64,),
    (29, "MegaCart 128 KB", 128,),
    (30, "MegaCart 256 KB", 256,),
    (31, "MegaCart 512 KB", 512,),
    (32, "MegaCart 1 MB", 1024,),
    (39, "Phoenix 8 KB", 8,),
    (46, "Blizzard 4 KB", 4,),
    (40, "Blizzard 16 KB", 16, 16, 0, 0x8000),
    (60, "Blizzard 32 KB", 32,),
    (41, "Atarimax 128 KB Flash", 128,),
    (42, "Atarimax 1 MB Flash", 1024,),
    (47, "AST 32 KB", 32,),
    (48, "Atrax SDX 64 KB", 64,),
    (49, "Atrax SDX 128 KB", 128,),
    (50, "Turbosoft 64 KB", 64,),
    (51, "Turbosoft 128 KB", 128,),
    (52, "Ultracart 32 KB", 32,),
    (53, "Low bank 8 KB", 8, 8, 0, 0x8000),
    (5,  "DB 32 KB", 32,),
    (54, "SIC! 128 KB", 128,),
    (55, "SIC! 256 KB", 256,),
    (56, "SIC! 512 KB", 512,),
    (61, "MegaMax 2 MB", 2048,),
    (62, "The!Cart 128 MB", 128*1024,),
    (63, "Flash MegaCart 4 MB", 4096,),
    (64, "MegaCart 2 MB", 2048,),
    (65, "The!Cart 32 MB", 32*1024,),
    (66, "The!Cart 64 MB", 64*1024,),
    (20, "Standard 4 KB 5200", 4, 4, 0, 0x8000),
    (19, "Standard 8 KB 5200", 8, 8, 0, 0x8000),
    (4,  "Standard 32 KB 5200", 32, 32, 0, 0x4000),
    (16, "One chip 16 KB 5200", 16,),
    (6,  "Two chip 16 KB 5200", 16,),
    (7,  "Bounty Bob 40 KB 5200", 40,),
]

known_cart_type_map = {c[0]:i for i, c in enumerate(known_cart_types)}


def get_known_carts():
    grouped = defaultdict(list)
    for c in known_cart_types[1:]:
        size = c[2]
        grouped[size].append(c)
    return grouped


def get_cart(cart_type):
    try:
        return known_cart_types[known_cart_type_map[cart_type]]
    except KeyError:
        raise errors.InvalidDiskImage("Unsupported cart type %d" % cart_type)


class A8CartHeader:
    # Atari Cart format described by https://sourceforge.net/p/atari800/source/ci/master/tree/DOC/cart.txt  NOTE: Big endian!
    format = np.dtype([
        ('magic', '|S4'),
        ('format', '>u4'),
        ('checksum', '>u4'),
        ('unused','>u4')
        ])
    nominal_length = format.itemsize
    file_format = "Cart"

    def __init__(self, bytes=None, create=False):
        self.image_size = 0
        self.cart_type = -1
        self.cart_name = ""
        self.cart_size = 0
        self.crc = 0
        self.unused = 0
        self.header_offset = 0
        self.num_banks = 0
        self.banks = []
        self.bank_size = 0
        self.bank_origin = 0
        self.main_size = 0
        self.main_offset = 0
        self.main_origin = 0
        self.possible_types = set()
        if create:
            self.header_offset = self.nominal_length
            self.check_size(0)
        if bytes is None:
            return

        if len(bytes) == 16:
            values = bytes.view(dtype=self.format)[0]
            if values[0] != b'CART':
                raise errors.InvalidCartHeader
            self.cart_type = int(values[1])
            self.crc = int(values[2])
            self.header_offset = self.nominal_length
            self.set_type(self.cart_type)
        else:
            raise errors.InvalidCartHeader

    def __str__(self):
        return "%s Cartridge (atari800 type=%d size=%d, %d banks, crc=%d)" % (self.cart_name, self.cart_type, self.cart_size, self.bank_size, self.crc)

    def __len__(self):
        return self.header_offset

    @property
    def valid(self):
        return self.cart_type != -1

    def calc_crc_from_data(self, data):
        self.crc = 0

    def to_array(self):
        raw = np.zeros([self.nominal_length], dtype=np.uint8)
        values = raw.view(dtype=self.format)[0]
        values[0] = b'CART'
        values[1] = self.cart_type
        values[2] = self.crc
        values[3] = 0
        return raw

    def set_type(self, cart_type):
        self.cart_type = cart_type
        c = get_cart(cart_type)
        self.cart_name = c[1]
        self.cart_size = c[2]
        self.main_size = self.cart_size
        if len(c) >= 6:
            self.main_size, self.main_offset, self.main_origin = c[3:6]
        if len(c) >= 9:
            self.banks = []
            self.bank_size, offset, self.bank_origin = c[6:9]
            s = self.cart_size - self.main_size
            while s > 0:
                self.banks.append(offset)
                offset += self.bank_size
                s -= self.bank_size

    def check_size(self, size):
        self.possible_types = set()
        k, r = divmod(size, 1024)
        if r == 0 or r == self.nominal_length:
            for i, t in enumerate(known_cart_types):
                valid_size = t[0]
                if k == valid_size:
                    self.possible_types.add(i)


class BaseAtariCartImage(DiskImageBase):
    def __str__(self):
        return str(self.header)

    def read_header(self):
        data = self.bytes[0:16]
        try:
            self.header = A8CartHeader(data)
        except errors.InvalidCartHeader:
            self.header = A8CartHeader()

    def strict_check(self):
        raise NotImplementedError

    def relaxed_check(self):
        if self.header.cart_type != self.cart_type:
            # force the header to be the specified cart type
            self.header = A8CartHeader()
            self.header.set_type(self.cart_type)
            self.check_size()

    def check_size(self):
        if not self.header.valid:
            return
        k, rem = divmod((len(self) - len(self.header)), 1024)
        c = get_cart(self.header.cart_type)
        log.debug("checking type=%d, k=%d, rem=%d for %s, %s" % (self.cart_type, k, rem, c[1], c[2]))
        if rem > 0:
            raise errors.InvalidDiskImage("Cart not multiple of 1K")
        if k != c[2]:
            raise errors.InvalidDiskImage("Image size %d doesn't match cart type %d size %d" % (k, self.cart_type, c[2]))

    def parse_segments(self):
        r = self.rawdata
        i = self.header.header_offset
        if i > 0:
            self.segments.append(ObjSegment(r[0:i], 0, 0, 0, i, name="Cart Header"))
        self.segments.extend(self.get_main_segment())
        self.segments.extend(self.get_banked_segments())

    def get_main_segment(self):
        r = self.rawdata
        start = self.header.header_offset + self.header.main_offset * 1024
        end = start + (self.header.main_size * 1024)
        s = ObjSegment(r[start:end], 0, 0, self.header.main_origin, name="Main Bank")
        return [s]

    def get_banked_segments(self):
        segments = []
        r = self.rawdata
        for i, offset in enumerate(self.header.banks):
            start = self.header.header_offset + offset * 1024
            end = start + (self.header.bank_size * 1024)
            s = ObjSegment(r[start:end], 0, 0, self.header.bank_origin, name="Bank #%d" % (i + 1))
            segments.append(s)
        return segments

    def create_emulator_boot_segment(self):
        h = self.header
        k, rem = divmod(len(self), 1024)
        if rem == 0:
            h.calc_crc_from_data(self.bytes)
            data_with_header = np.empty(len(self) + h.nominal_length, dtype=np.uint8)
            data_with_header[0:h.nominal_length] = h.to_array()
            data_with_header[h.nominal_length:] = self.bytes
            r = SegmentData(data_with_header)
        else:
            r = self.rawdata
        s = ObjSegment(r, 0, 0, self.header.main_origin, name="Cart image")
        return s


class AtariCartImage(BaseAtariCartImage):
    def __init__(self, rawdata, cart_type, filename=""):
        c = get_cart(cart_type)
        self.cart_type = cart_type
        DiskImageBase.__init__(self, rawdata, filename)

    def strict_check(self):
        if not self.header.valid:
            raise errors.InvalidDiskImage("Missing cart header")
        if self.header.cart_type != self.cart_type:
            raise errors.InvalidDiskImage("Cart type doesn't match type defined in header")


class Atari8bitCartImage(AtariCartImage):
    def strict_check(self):
        if "5200" in self.header.cart_name:
            raise errors.InvalidDiskImage("5200 Carts don't work in the home computers.")
        AtariCartImage.strict_check(self)


class Atari5200CartImage(AtariCartImage):
    def strict_check(self):
        if "5200" not in self.header.cart_name:
            raise errors.InvalidDiskImage("Home computer carts don't work in the 5200.")
        AtariCartImage.strict_check(self)


def add_cart_header(bytes):
    header = A8CartHeader(create=True)
    header.check_size(len(bytes))
    hlen = len(header)
    data = np.empty([hlen + len(bytes)], dtype=np.uint8)
    data[0:hlen] = header.to_array()
    data[hlen:] = bytes
    return data


class RomImage(DiskImageBase):
    def __str__(self):
        return f"{len(self.rawdata) // 1024}k ROM image"

    def read_header(self):
        self.header = A8CartHeader()

    def strict_check(self):
        self.check_size()

    def check_size(self):
        size = len(self)
        if (size & (size - 1)) != 0:
            raise errors.InvalidDiskImage("ROM image not a power of 2")

    def parse_segments(self):
        r = self.rawdata
        s = ObjSegment(r, 0, 0, self.header.main_origin, name="Main Bank")
        self.segments = [s]

    def create_emulator_boot_segment(self):
        s = self.segments[0]
        if s.origin == 0:
            return None
        return s


class Atari2600CartImage(RomImage):
    def __str__(self):
        return f"{len(self.rawdata) // 1024}k Atari 2600 Cartridge"


class Atari2600StarpathImage(RomImage):
    def __str__(self):
        return f"{len(self.rawdata) // 1024}k Atari 2600 Starpath Cassette"


class VectrexCartImage(RomImage):
    def __str__(self):
        return f"{len(self.rawdata) // 1024}k Vectrex Cartridge"



import gzip
import io
import tarfile
import time

import numpy as np

from .. import errors
from ..archiver import Archiver

import logging
log = logging.getLogger(__name__)


class TarArchiver(Archiver):
    archive_type = "tar"

    def iter_archive(self, basename, byte_data):
        byte_data = io.BytesIO(byte_data)
        try:
            with tarfile.open(None, "r", byte_data) as zf:
                for item in zf.getmembers():
                    log.debug(f"tarinfo item: {item}")
                    item_data = np.frombuffer(zf.extractfile(item).read(), dtype=np.uint8)
                    yield item.name, item_data
        except:
            raise errors.InvalidArchiver("Not a tar file")

    def get_tarinfo(self, container):
        tarinfo = tarfile.TarInfo(container.pathname)
        tarinfo.mtime = time.time()
        tarinfo.uid = 400  # easter egg!
        tarinfo.gid = 800
        return tarinfo

    def pack_data(self, fh, containers, skip_missing_compressors=False):
        with tarfile.open(None, 'w', fh) as zf:
            for c in containers:
                byte_data = c.calc_packed_bytes(skip_missing_compressors)
                d = io.BytesIO(byte_data)
                tarinfo = self.get_tarinfo(c)
                tarinfo.size = len(byte_data)
                zf.addfile(tarinfo, d)



import hashlib
import inspect
import pkg_resources

import numpy as np

from . import errors
from . import style_bits
from .segment import Segment
from .utils import to_numpy, to_numpy_list, uuid
from . import filesystem
from .file_type import guess_file_type
from .signature import guess_signature_from_container

import logging
log = logging.getLogger(__name__)


class Media(Segment):
    """Base class for what is typically the root segment in a Container,
    describing the type of media the data represents: floppy disk image,
    cassette image, cartridge, etc.

    Media can contain a filesystem, which is used to further subdivide the data
    from the container into segments to represent each logical section in the
    filesystem (i.e. boot sectors, VTOC, directory structure, files, etc.)
    """
    ui_name = "Unknown media"

    extra_serializable_attributes = []

    def __init__(self, container, signature=None, force=False):
        container.header = self.calc_header(container)
        size = len(container) - container.header_length
        Segment.__init__(self, container, container.header_length, name=self.ui_name, length=size)
        self.signature = signature
        if container.header is not None:
            self.check_header(container.header)
        try:
            self.check_media_size()
        except errors.PossibleCandidateMedia:
            if not force:
                raise
        self.check_magic()

    def __str__(self):
        desc = f"{self.ui_name}, size={len(self)}"
        if self.filesystem is not None:
            desc += f", filesystem={self.filesystem.ui_name}"
        return desc

    @property
    def filesystem(self):
        return self.container.filesystem

    #### initialization

    def calc_header(self, container):
        """Subclasses should override this method to verify the integrity of
        any header information, if any.

        If a header does exist, the subclass must return a `Segment` containing
        the header.

        Subclasses should raise the appropriate MediaError if the data is
        incompatible with this media type.
        """
        pass

    def check_header(self, header):
        """Subclasses should override this method to verify that header data is
        consistent with the media segment.

        Subclasses should raise the appropriate MediaError if the header and
        data are inconsistent.
        """
        pass

    def check_media_size(self):
        """Subclasses should override this method to verify that the payload
        portion of the media (i.e. everything after the header) can be stored
        in this media.

        Subclasses should raise the appropriate MediaError if the data is
        incompatible with this media type.
        """
        pass

    def check_magic(self):
        """Subclasses should override this method if there is some "magic"
        values that can identify (or rule out) this disk image class as a
        canditate.

        Subclasses should raise the appropriate MediaError if the data is
        incompatible with this media type.
        """
        pass


class DiskImage(Media):
    ui_name = "Disk Image"
    sector_size = 128
    expected_size = 0
    starting_sector_label = 1
    first_sector = 1

    extra_serializable_attributes = ['num_sectors']

    def init_empty(self):
        super().init_empty()
        self.num_sectors = 0

    # def __str__(self):
    #     return f"{self.ui_name}, size={len(self)} ({self.num_sectors}x{self.sector_size}B)"

    #### verification

    def check_media_size(self):
        self.check_disk_size()
        self.num_sectors = self.calc_num_sectors()

    def check_disk_size(self):
        size = len(self)
        if size != self.expected_size:
            raise errors.InvalidMediaSize(f"{self.ui_name} expects size {self.expected_size}; found {size}")

    def calc_num_sectors(self):
        size = len(self)
        if size % self.sector_size != 0:
            raise errors.InvalidMediaSize(f"{self.ui_name} requires integer number of sectors")
        return size // self.sector_size

    #### sector operations

    def get_ui_name_at_index(self, index, lower_case=True):
        sector, byte = divmod(index, self.sector_size)
        if lower_case:
            return "s%03d:%02x" % (sector + self.first_sector, byte)
        return "s%03d:%02X" % (sector + self.first_sector, byte)

    def is_sector_valid(self, sector):
        return (self.num_sectors < 0) or (sector >= self.starting_sector_label and sector < (self.num_sectors + self.starting_sector_label))

    def get_index_of_sector(self, sector):
        if not self.is_sector_valid(sector):
            raise errors.InvalidSectorNumber("Sector %d out of range" % sector)
        pos = (sector - self.starting_sector_label) * self.sector_size
        return pos, self.sector_size

    def get_contiguous_sectors_offsets(self, start, count=1):
        index, _ = self.get_index_of_sector(start)
        last, size = self.get_index_of_sector(start + count - 1)
        return index, last + size - index

    def get_contiguous_sectors(self, start, count=1):
        start, size = self.get_contiguous_sectors_offsets(start, count)
        return Segment(self, start, length=size)

    def get_sector_list_offsets(self, sector_numbers, sector_size_override=None):
        if sector_size_override is not None:
            sector_size = sector_size_override
        else:
            sector_size = self.sector_size
        offsets = np.empty(len(sector_numbers) * sector_size, dtype=np.uint32)
        i = 0
        for num in sector_numbers:
            index, size = self.get_index_of_sector(num)
            size = min(size, sector_size)
            offsets[i:i+size] = np.arange(index, index + size)
            i += size
        if i < len(offsets):
            # can happen in Atari DD disks; 1st 3 sectors are 128 bytes, not
            # 256
            offsets = np.copy(offsets[0:i])
        return offsets

    def get_sector_list(self, sector_numbers):
        offsets = self.get_sector_list_offsets(sector_numbers)
        return Segment(self, offsets)

    def iter_sectors(self):
        i = self.starting_sector_label
        while self.is_sector_valid(i):
            pos, size = self.get_index_of_sector(i)
            yield i, pos, size
            i += 1


class CartImage(Media):
    ui_name = "Cart Image"
    platform = "emulator_name"
    cart_type = 0
    expected_size = 0

    extra_serializable_attributes = ['cart_type']

    def __str__(self):
        desc = f"{self.ui_name}, size={self.kb}K, {self.platform}, cart_type={self.cart_type}"
        if self.filesystem is not None:
            desc += f", filesystem={self.filesystem.ui_name}"
        return desc

    def check_media_size(self):
        size = len(self)
        self.kb, rem = divmod(size, 1024)
        if rem > 0:
            raise errors.InvalidMediaSize("Cart not multiple of 1K")
        if self.expected_size == 0:
            raise errors.PossibleCandidateMedia(f"Possible cart image, but unable to identify specifically")
        if size != self.expected_size:
            raise errors.InvalidMediaSize(f"{self.ui_name} expects size {self.expected_size}; found {size}")


ignore_base_class_media_types = set([DiskImage, CartImage])

_media_types = None

def _find_media_types():
    media_types = []
    for entry_point in pkg_resources.iter_entry_points('atrip.media_types'):
        mod = entry_point.load()
        log.debug(f"find_media_type: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Media in obj.__mro__[1:] and obj not in ignore_base_class_media_types:
                log.debug(f"find_media_types:   found media_type class {name}")
                media_types.append(obj)
    return media_types

def find_media_types():
    global _media_types

    if _media_types is None:
        _media_types = _find_media_types()
    return _media_types

def guess_media_type(container):
    signature = guess_signature_from_container(container)
    if signature:
        log.info(f"found signature {signature}")
    possibilities = []
    for m in find_media_types():
        log.debug(f"trying media_type {m.ui_name}")
        try:
            found = m(container, signature)
        except errors.PossibleCandidateMedia as e:
            log.info(f"might be media_type: {m}")
            possibilities.append(m)
        except errors.MediaError as e:
            log.debug(f"found error: {e}")
            continue
        else:
            log.info(f"found media_type {m.ui_name}")
            return found
    if possibilities:
        log.info(f"{len(possibilities)} possible media types")
    for m in possibilities:
        try:
            found = m(container, signature, True)
        except errors.MediaError as e:
            log.info(f"not choosing media type {m.ui_name}: {e}")
            continue
        else:
            log.info(f"choosing media type {m.ui_name}.")
            return found
    log.info(f"No recognized media type.")
    m = Media
    return m(container, signature, True)



# flags

# udis cpu opcode flags
udis_opcode_flag_branch = 1
udis_opcode_flag_jump = 2
udis_opcode_flag_return = 4
udis_opcode_flag_label = 8
udis_opcode_flag_data_bytes = 16
udis_opcode_flag_store = 32  # is this a store operation? default is load
udis_opcode_flag_undoc = 128

# flags
FLAG_BRANCH_TAKEN = 1
FLAG_BRANCH_NOT_TAKEN = 2
FLAG_REPEATED_BYTES = 3
FLAG_REG_A = 4
FLAG_REG_X = 5
FLAG_REG_Y = 6
FLAG_LOAD_A_FROM_MEMORY = 7
FLAG_LOAD_X_FROM_MEMORY = 8
FLAG_LOAD_Y_FROM_MEMORY = 9
FLAG_MEMORY_ALTER = 10
FLAG_MEMORY_READ_ALTER_A = 11 
FLAG_PEEK_MEMORY = 12
FLAG_PULL_A = 13
FLAG_PULL_SR = 14
FLAG_PUSH_A = 15
FLAG_PUSH_SR = 16
FLAG_RTI = 17
FLAG_RTS = 18
FLAG_STORE_A_IN_MEMORY = 19
FLAG_STORE_X_IN_MEMORY = 20
FLAG_STORE_Y_IN_MEMORY = 21
FLAG_RESULT_MASK = 0x3f
FLAG_TARGET_ADDR = 64
FLAG_REG_SR = 128

# disassembler types
DISASM_DATA = 0
DISASM_6502 = 10
DISASM_65816 = 12
DISASM_65C02 = 13
DISASM_6800 = 14
DISASM_6809 = 15
DISASM_6811 = 16
DISASM_8051 = 17
DISASM_8080 = 18
DISASM_Z80 = 19
DISASM_ANTIC_DL = 30
DISASM_JUMPMAN_HARVEST = 31
DISASM_JUMPMAN_LEVEL = 32

# types 128-191 are for history entries that have result entries
DISASM_6502_HISTORY = 128
DISASM_6502_HISTORY_RESULT = 129
DISASM_ATARI800_HISTORY = 130
DISASM_ATARI800_HISTORY_RESULT = 131
DISASM_NEXT_INSTRUCTION = 132
DISASM_NEXT_INSTRUCTION_RESULT = 133

# types 192-254 don't have results
DISASM_FRAME_START = 192
DISASM_FRAME_END = 193
DISASM_ATARI800_VBI_START = 194
DISASM_ATARI800_VBI_END = 195
DISASM_ATARI800_DLI_START = 196
DISASM_ATARI800_DLI_END = 197

DISASM_BREAKPOINT = 253
DISASM_USER_DEFINED = 254
DISASM_UNKNOWN = 255

udis_opcode_flag_map = {
    udis_opcode_flag_branch: "FLAG_BRANCH",
    udis_opcode_flag_jump: "FLAG_JUMP",
    udis_opcode_flag_return: "FLAG_RTS",
    udis_opcode_flag_label: "FLAG_LABEL",
    udis_opcode_flag_undoc: "FLAG_UNDOC",
}

pcr = 256 # for all currently defined CPUs: any instruction that uses this flag indicates a branch instruction
und = 512
z80bit = 1024
lbl = 8
comment = 2048



import numpy as np

from . import errors
from .segments import SegmentData
from .diskimages import BaseHeader, DiskImageBase

import logging
log = logging.getLogger(__name__)


class StandardDeliveryHeader(BaseHeader):
    file_format = "Apple ][ Standard Delivery"

    def __init__(self, bytes=None, sector_size=256, create=False):
        BaseHeader.__init__(self, sector_size, create=create)
        if bytes is None:
            return

        data = bytes[0:5]
        if np.all(data == (0x01, 0xa8, 0xee, 0x06, 0x08)):
            log.debug("Found 48k loader")
        else:
            raise errors.InvalidDiskImage("No %s boot header" % self.file_format)

    def __str__(self):
        return "Standard Delivery Boot Disk (size=%d (%dx%dB)" % (self.file_format, self.image_size, self.max_sectors, self.sector_size)

    def check_size(self, size):
        if size != 143360:
            raise errors.InvalidDiskImage("Incorrect size for Standard Delivery image")
        self.image_size = size
        self.tracks_per_disk = 35
        self.sectors_per_track = 16
        self.max_sectors = self.tracks_per_disk * self.sectors_per_track


class StandardDeliveryImage(DiskImageBase):
    def __str__(self):
        return str(self.header)

    def read_header(self):
        self.header = StandardDeliveryHeader(self.bytes[0:256])

    @classmethod
    def new_header(cls, diskimage, format="DSK"):
        if format.lower() == "dsk":
            header = StandardDeliveryHeader(create=True)
            header.check_size(diskimage.size)
        else:
            raise RuntimeError("Unknown header type %s" % format)
        return header

    def check_size(self):
        pass

    def get_boot_sector_info(self):
        pass

    def get_vtoc(self):
        pass

    def get_directory(self, directory=None):
        pass

    @classmethod
    def create_boot_image(cls, segments, run_addr=None):
        raw = SegmentData(np.zeros([143360], dtype=np.uint8))
        dsk = cls(raw, create=True)
        if run_addr is None:
            run_addr = segments[0].origin

        chunks = []

        for s in segments:
            # find size in 256 byte chunks that start on a page boundary
            # since the loader only deals with page boundaries
            origin = s.origin
            chunk_start, padding = divmod(origin, 256)
            if (chunk_start == 0x20 or chunk_start == 0x40) and padding == 1:
                show_hgr = False
                padding = 0
            else:
                show_hgr = True
            size = ((len(s) + padding + 255) // 256) * 256
            chunk = np.zeros([size], dtype=np.uint8)
            chunk[padding:padding + len(s)] = s[:]
            chunks.append((chunk_start, chunk, show_hgr))
            print("segment: %s, pages=%d" % (str(s), len(chunk) // 256))
            log.debug(" last chunk=%s" % str(chunks[-1]))

        # break up the chunks into sectors

        # NOTE: fstbt implied that the sector order was staggered, but in
        # AppleWin, by trial and error, works with the following order

        # index = 1  # on the first track, sector 0 is reserved for boot sector
        # sector_order = [0, 2, 4, 6, 8, 10, 12, 14, 1, 3, 5, 7, 9, 11, 13, 15]

        index = 1
        sector_order = [0, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 15]
        track = 0
        count = 0

        sector_list = []
        address_list = [0xd1]

        boot_sector = dsk.header.create_sector()
        boot_sector.sector_num = 0
        boot_sector.track_num = 1
        sector_list.append(boot_sector)
        first_page_1 = True

        for chunk_start, chunk_data, show_hgr in chunks:
            count = len(chunk_data) // 256
            if chunk_start == 0x20 and count == 32 and first_page_1:
                # Assume this is an HGR screen, use interesting load effect,
                # not the usual venetian blind
                chunk_hi = [0x20, 0x24, 0x28, 0x2c, 0x30, 0x34, 0x38, 0x3c, 0x21, 0x25, 0x29, 0x2d, 0x31, 0x35, 0x39, 0x3d, 0x22, 0x26, 0x2a, 0x2e, 0x32, 0x36, 0x3a, 0x3e, 0x23, 0x27, 0x2b, 0x2f, 0x33, 0x37, 0x3b, 0x3f]
                address_order = chunk_hi
            else:
                chunk_hi = range(chunk_start, chunk_start + count)
                if len(chunk_hi) > 2:
                    address_order = []
                    for n in range(0, count, 32):
                        subset = chunk_hi[n:n+32]
                        if len(subset) > 2:
                            address_order.extend([chunk_hi[n], 0xe0 + len(subset) - 1])
                        else:
                            address_order.extend(subset)
                else:
                    address_order = chunk_hi

            for n in range(count):
                i = (chunk_hi[n] - chunk_start) * 256
                sector = dsk.header.create_sector(chunk_data[i:i+256])
                sector.sector_num = dsk.header.sector_from_track(track, sector_order[index])
                count += 1
                #sector.sector_num = count
                sector_list.append(sector)
                # sector.data[0] = sector.sector_num
                # sector.data[1] = hi
                # sector.data[2:16] = 0xff
                log.debug("%s at %02x00: %s ..." % (sector_list[-1], address_list[-1], " ".join(["%02x" % h for h in chunk_data[i:i + 16]])))
                index += 1
                if index >= len(sector_order):
                    index = 0
                    track += 1
            address_list.extend(address_order)
            if show_hgr:
                if chunk_start == 0x40:
                    address_list.append(0xd2)
                elif chunk_start == 0x20:
                    if not first_page_1:
                        address_list.append(0xd1)
                    first_page_1 = False

        print("fstbt commands: %s" % ", ".join(["%02x" % i for i in address_list]))
        boot_code = get_fstbt_code(boot_sector.data, address_list, run_addr)

        dsk.write_sector_list(sector_list)

        return dsk

from . fstbt import fstbt

def get_fstbt_code(data, address_list, run_addr):
    pointer = len(fstbt)
    data[0:pointer] = np.frombuffer(fstbt, dtype=np.uint8)
    hi, lo = divmod(run_addr, 256)
    data[pointer:pointer + 2] = (lo, hi)
    address_list.append(0xc0)  # last sector flag
    data[pointer + 2:pointer + 2 + len(address_list)] = address_list



import sys
from collections import namedtuple

import numpy as np
import jsonpickle

from atrip import Collection, Container, Segment, errors

from sawx.document import SawxDocument
from sawx.utils.nputil import to_numpy
from sawx.events import EventHandler

from atrip.disassembler import DisassemblyConfig, valid_cpu_ids, cpu_name_to_id
from .utils.templateutil import load_memory_map

import logging
log = logging.getLogger(__name__)


class DiskImageDocument(SawxDocument):
    """Document for atrcopy-parsed segmented files

    Events:

    During high framerate operations, some panels may not need to be updated
    very frequently so only those panels that absolutely need it will get a
    high priority refresh event. Others will get lower priority. An integer
    should be passed as this event's data; it is up to the viewers to decide
    what the value means. Confusingly, high priority levels are lower
    numbers! This could mean the number of frames to skip, but it's up to the
    viewers, really.

    """
    json_expand_keywords = {
        'linked bases': 2,
        'viewers': 2,
    }

    session_save_file_extension = ".omnivore"

    def __init__(self, file_metadata):
        self.document_memory_map = {}
        self._cpu = "6502"
        self._disassembler = None
        self._operating_system = "atari800"
        self._machine_labels = None
        self.skip_frames_on_boot = 0
        self.pause_emulator_on_boot = False
        SawxDocument.__init__(self, file_metadata)

        self.cpu_changed_event = EventHandler(self)

        # default emulator class, if the user selects something different than
        # the normal default. This is usually None, which means that omnivore
        # will chose the best emulator based on the type of this segment
        self.emulator_class_override = None

    @property
    def can_resize(self):
        return self.segments and self.collection.can_resize

    @property
    def labels(self):
        return self.machine_labels#.labels

    @property
    def cpu(self):
        return self._cpu

    @cpu.setter
    def cpu(self, value):
        self._cpu = value
        self._disassembler = None
        self.cpu_changed_event(value)

    @property
    def operating_system(self):
        return self._operating_system

    @operating_system.setter
    def operating_system(self, value):
        self._operating_system = value
        self._machine_labels = None
        self.cpu_changed_event(value)

    @property
    def machine_labels(self):
        if self._machine_labels is None:
            self._machine_labels = load_memory_map(self._operating_system)
        return self._machine_labels

    @property
    def disassembler(self):
        if self._disassembler is None:
            default_cpu = cpu_name_to_id.get(self._cpu, 0)
            driver = DisassemblyConfig(default_cpu)
            # driver.register_parser(self._cpu, 0)
            self._disassembler = driver
        return self._disassembler

    #### object methods

    def __str__(self):
        lines = []
        lines.append(f"DiskImageDocument: uuid={self.uuid}, {self.collection}")
        if log.isEnabledFor(logging.DEBUG):
            lines.extend(self.collection.verbose_info.splitlines())
        return "\n---- ".join(lines)

    def __len__(self):
        # different that superclass; reports number of containers in collection
        return len(self.collection)

    #### loaders

    def load(self, file_metadata):
        log.debug(f"load: file_metadata={file_metadata}")
        collection = file_metadata["atrip_collection"]
        self.load_collection(collection, file_metadata)

    def load_collection(self, collection, file_metadata):
        self.collection = collection
        log.debug(f"load: found collection {self.collection}")
        self.file_metadata = file_metadata
        self.load_session()
        self.segments = list(self.collection.iter_segments())
        self.user_segments = []

    def calc_raw_data_to_save(self):
        return self.collection.calc_compressed_data()

    #### serialization methods

    def serialize_session(self, s):
        SawxDocument.serialize_session(self, s)
        s2 = {}
        self.collection.serialize_session(s2)
        s["atrip_session"] = s2

        s["serialized user segments"] = list(self.user_segments)
        s["document memory map"] = sorted([list(i) for i in list(self.document_memory_map.items())])  # save as list of pairs because json doesn't allow int keys for dict
        s["skip_frames_on_boot"] = self.skip_frames_on_boot
        s["pause_emulator_on_boot"] = self.pause_emulator_on_boot

    def restore_session(self, e):
        SawxDocument.restore_session(self, e)
        s2 = e.get("atrip_session", None)
        if s2 is not None:
            self.collection.restore_session(s2)
        if 'user segments' in e:
            # Segment objects created by the utils.extra_metadata module
            for s in e['user segments']:
                self.add_user_segment(s, replace=True)
        if 'serialized user segments' in e:
            log.warning("found serialized user segments from Omnivore 1.0. These will not be restored.")
        if 'document memory map' in e:
            self.document_memory_map = dict(e['document memory map'])
        if 'skip_frames_on_boot' in e:
            self.skip_frames_on_boot = e['skip_frames_on_boot']
        if 'pause_emulator_on_boot' in e:
            self.pause_emulator_on_boot = e['pause_emulator_on_boot']

    #### convenience methods

    def calc_layout_template_name(self, task_id):
        return "%s.default_layout" % task_id

    def find_matching_segment(self, segment):
        for s in self.segments:
            if len(s) == len(segment) and s.origin == segment.origin and s.name == segment.name:
                return True
        return False

    def find_matching_user_segment(self, segment):
        for s in self.user_segments:
            if len(s) == len(segment) and s.origin == segment.origin and s.name == segment.name:
                return s
        return None

    def find_segment_index(self, segment):
        try:
            return self.segments.index(segment)
        except ValueError:
            for i, s in enumerate(self.segments):
                if s.name == segment or s.uuid == segment:
                    return i
        return -1

    def find_segment_by_name(self, name):
        """Assuming segments had a origin param, find first segment that
        has addr as a valid address
        """
        for i, s in enumerate(self.segments):
            if s.name == name:
                return s
        return None

    def find_segments_in_range(self, addr):
        """Assuming segments had a origin param, find first segment that
        has addr as a valid address
        """
        found = []
        for i, s in enumerate(self.segments):
            if s.origin > 0 and addr >= s.origin and addr < (s.origin + len(s)):
                found.append((i, s, addr - s.origin))
        return found

    def find_segments_with_raw_index(self, raw_index):
        """Find all segments that contain the specified raw index

        The raw index points to a specific byte, so this will return all
        segments that have a view of this byte. This function ignores the
        segment start address because different views may have different start
        addresses; to find segments that contain a specific address, use
        find_segment_in_range.
        """
        found = []
        for i, s in enumerate(self.segments):
            try:
                index = s.get_index_from_base_index(raw_index)
                found.append((i, s, index))
            except IndexError:
                pass
        return found

    ##### Initial viewer defaults

    def find_initial_visible_segment(self, linked_base, default=None):
        """Hook for subclasses to force a particular segment to be viewed on
        document load. Used in emulators to show the main memory, which is not
        usually the first segment in the list.

        By default, it does show the first segment.
        """
        segment = self.collection.containers[0].segments[0]
        linked_base.find_segment(segment, refresh=False)

    #### Baseline document for comparisons

    def init_baseline(self, metadata, raw_data):
        d = DiskImageDocument(metadata=metadata, raw_data=raw_data)
        d.parse_segments([])
        self.baseline_document = d

    def del_baseline(self):
        self.baseline_document = None

    def update_baseline(self):
        if self.baseline_document is not None:
            self.change_count += 1
            self.collection.compare_segment(self.baseline_document.collection)

    def clear_baseline(self):
        self.change_count += 1
        self.collection.clear_style_bits(diff=True)

    @property
    def has_baseline(self):
        return self.baseline_document is not None

    @classmethod
    def create_from_segments(cls, root, user_segments):
        doc = cls(raw_data=root.data, style=root.style)
        Parser = namedtuple("Parser", ['segments'])
        segs = [root]
        p = Parser(segments=segs)
        doc.user_segments = list(user_segments)
        doc.set_segments(p)
        return doc

    #### file recognition

    @classmethod
    def can_load_file_exact(cls, file_metadata):
        return "atrip_identified" in file_metadata
 
    @classmethod
    def can_load_file_generic(cls, file_metadata):
        mime_type = file_metadata['mime']
        return mime_type == "application/octet-stream"



import io

from .. import errors
from ..compressor import Compressor


class UnixCompressor(Compressor):
    compression_algorithm = "unix compress"

    def calc_unpacked_data(self, byte_data):
        try:
            unpacked = unlzw(byte_data)
        except ValueError as e:
            raise errors.InvalidAlgorithm(e)
        return unpacked


def unlzw(data):
    """
    This function was adapted for Python from Mark Adler's C implementation
    https://github.com/umeat/unlzw version 0.1.0

    Decompress compressed data generated by the Unix compress utility (LZW
    compression, files with .Z suffix). Input can be given as any type which
    can be 'converted' to a bytearray (e.g. string, or bytearray). Returns
    decompressed data as string, or raises error.

    Written by Brandon Owen, May 2016, brandon.owen@hotmail.com
    Adapted from original work by Mark Adler - orginal copyright notice below

    Copyright (C) 2014, 2015 Mark Adler
    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.
    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:
    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.
    Mark Adler
    madler@alumni.caltech.edu
    """

    # Convert input data stream to byte array, and get length of that array
    try:
        ba_in = bytearray(data)
    except ValueError:
        raise TypeError("Unable to convert inputted data to bytearray")

    inlen = len(ba_in)
    prefix = [None] * 65536  # index to LZW prefix string
    suffix = [None] * 65536  # one-character LZW suffix

    # Process header
    if inlen < 3:
        raise ValueError(
            "Invalid Input: Length of input too short for processing")

    if (ba_in[0] != 0x1f) or (ba_in[1] != 0x9d):
        raise ValueError(
            "Invalid Header Flags Byte: Incorrect magic bytes")

    flags = ba_in[2]

    if flags & 0x60:
        raise ValueError(
            "Invalid Header Flags Byte: Flag byte contains invalid data")

    max_ = flags & 0x1f
    if (max_ < 9) or (max_ > 16):
        raise ValueError(
            "Invalid Header Flags Byte: Max code size bits out of range")

    if (max_ == 9):
        max_ = 10  # 9 doesn't really mean 9

    flags &= 0x80  # true if block compressed

    # Clear table, start at nine bits per symbol
    bits = 9
    mask = 0x1ff
    end = 256 if flags else 255

    # Ensure stream is initially valid
    if inlen == 3:
        return 0  # zero-length input is permitted
    if inlen == 4:  # a partial code is not okay
        raise ValueError("Invalid Data: Stream ended in the middle of a code")

    # Set up: get the first 9-bit code, which is the first decompressed byte,
    # but don't create a table entry until the next code
    buf = ba_in[3]
    buf += ba_in[4] << 8
    final = prev = buf & mask  # code
    buf >>= bits
    left = 16 - bits
    if prev > 255:
        raise ValueError("Invalid Data: First code must be a literal")

    # We have output - allocate and set up an output buffer with first byte
    put = [final]

    # Decode codes
    mark = 3  # start of compressed data
    nxt = 5  # consumed five bytes so far
    while nxt < inlen:
        # If the table will be full after this, increment the code size
        if (end >= mask) and (bits < max_):
            # Flush unused input bits and bytes to next 8*bits bit boundary
            # (this is a vestigial aspect of the compressed data format
            # derived from an implementation that made use of a special VAX
            # machine instruction!)
            rem = (nxt - mark) % bits

            if (rem):
                rem = bits - rem
                if rem >= inlen - nxt:
                    break
                nxt += rem

            buf = 0
            left = 0

            # mark this new location for computing the next flush
            mark = nxt

            # increment the number of bits per symbol
            bits += 1
            mask <<= 1
            mask += 1

        # Get a code of bits bits
        buf += ba_in[nxt] << left
        nxt += 1
        left += 8
        if left < bits:
            if nxt == inlen:
                raise ValueError(
                    "Invalid Data: Stream ended in the middle of a code")
            buf += ba_in[nxt] << left
            nxt += 1
            left += 8
        code = buf & mask
        buf >>= bits
        left -= bits

        # process clear code (256)
        if (code == 256) and flags:
            # Flush unused input bits and bytes to next 8*bits bit boundary
            rem = (nxt - mark) % bits
            if rem:
                rem = bits - rem
                if rem > inlen - nxt:
                    break
                nxt += rem
            buf = 0
            left = 0

            # Mark this location for computing the next flush
            mark = nxt

            # Go back to nine bits per symbol
            bits = 9  # initialize bits and mask
            mask = 0x1ff
            end = 255  # empty table
            continue  # get next code

        # Process LZW code
        temp = code  # save the current code
        stack = []  # buffer for reversed match - empty stack

        # Special code to reuse last match
        if code > end:
            # Be picky on the allowed code here, and make sure that the
            # code we drop through (prev) will be a valid index so that
            # random input does not cause an exception
            if (code != end + 1) or (prev > end):
                raise ValueError("Invalid Data: Invalid code detected")
            stack.append(final)
            code = prev

        # Walk through linked list to generate output in reverse order
        while code >= 256:
            stack.append(suffix[code])
            code = prefix[code]

        stack.append(code)
        final = code

        # Link new table entry
        if end < mask:
            end += 1
            prefix[end] = prev
            suffix[end] = final

        # Set previous code for next iteration
        prev = temp

        # Write stack to output in forward order
        put += stack[::-1]

    # Return the decompressed data as string
    return bytes(bytearray(put))



"""Byte modification actions
"""
import os
import sys

import wx

from sawx.ui import dialogs
from atrip.char_mapping import font_mapping_list
from atrip.machines import atari8bit

from ..action import ViewerAction, ViewerListAction, ViewerRadioListAction
from .. import commands
from ..arch.bitmap_renderers import bitmap_renderer_list
from ..arch.fonts import font_list, font_groups, prompt_for_font_from_group
from ..arch.font_renderers import font_renderer_list
from ..arch.ui.antic_colors import AnticColorDialog
from ..viewer import find_viewer_class_by_name, get_viewers

import logging
log = logging.getLogger(__name__)


class view_width(ViewerAction):
    """Set the number of items per row of the bitmap display. The width can
    mean different things for different viewers (i.e. bitmap widths are in
    byte_values, not pixels), so prompt is based on the viewer.
    """
    name = "Width..."

    def calc_enabled(self, action_key):
        state = self.viewer.has_width
        return state

    def perform(self, action_key):
        v = self.viewer
        val = dialogs.prompt_for_dec(v.control, 'Enter new %s' % v.width_text, 'Set Width', v.width)
        if val is not None and val > 0:
            v.set_width(val)


class view_zoom(ViewerAction):
    """Set the zoom factor of the viewer, if applicable. This is an integer
    value greater than zero that scales the display size of each item.
    """
    name = "Zoom..."

    def calc_enabled(self, action_key):
        state = self.viewer.has_zoom
        return state

    def perform(self, event):
        v = self.viewer
        val = dialogs.prompt_for_dec(v.control, 'Enter new %s' % v.zoom_text, 'Set Zoom', v.zoom)
        if val is not None and val > 0:
            v.set_zoom(val)


class ColorAction(ViewerAction):
    def calc_enabled(self, action_key):
        state = self.viewer.has_colors
        return state


class view_ask_colors(ColorAction):
    """Open a window to choose the color palette from the available colors
    of the ANTIC processor.
    """
    name = 'Use ANTIC Colors...'

    def perform(self, action_key):
        v = self.viewer
        dlg = AnticColorDialog(v.control, v.antic_color_registers, v.linked_base.cached_preferences)
        if dlg.ShowModal() == wx.ID_OK:
            v.antic_color_registers = dlg.colors


class view_antic_powerup_colors(ColorAction):
    """Changes the color palette to {name}
    """
    name = 'ANTIC Powerup Colors'

    def perform(self, action_key):
        self.viewer.antic_color_registers = atari8bit.powerup_colors()


class view_color_standards(ViewerRadioListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_colors

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        return atari8bit.color_standard_list

    def calc_checked_list_item(self, action_key, index, item):
        return self.viewer.color_standard_name == item.name

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.viewer.color_standard_name = item.name


class view_bitmap_renderers(ViewerRadioListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_bitmap

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        return bitmap_renderer_list

    def calc_checked_list_item(self, action_key, index, item):
        return self.viewer.bitmap_renderer_name == item.name

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.viewer.bitmap_renderer_name = item.name


class view_font_renderers(ViewerRadioListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_font

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        return font_renderer_list

    def calc_checked_list_item(self, action_key, index, item):
        return self.viewer.font_renderer_name == item.name

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.viewer.font_renderer_name = item.name


class view_font_mappings(ViewerRadioListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_font

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        return font_mapping_list

    def calc_checked_list_item(self, action_key, index, item):
        return self.viewer.font_mapping_name == item.name

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.viewer.font_mapping_name = item.name


class view_fonts(ViewerRadioListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_font

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item)

    def calc_list_items(self):
        return font_list

    def calc_checked_list_item(self, action_key, index, item):
        return self.viewer.antic_font_uuid == item["uuid"]

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.viewer.antic_font_data = item


class view_font_groups(ViewerListAction):
    def calc_enabled(self, action_key):
        return self.viewer.has_font

    def calc_name(self, action_key):
        item = self.get_item(action_key)
        return str(item) + "..."

    def calc_list_items(self):
        return sorted(font_groups.keys())

    def callback(self, font):
        self.viewer.antic_font_data = font

    def perform(self, action_key):
        item = self.get_item(action_key)
        current_font = self.viewer.antic_font_data
        font = prompt_for_font_from_group(self.editor.frame, item, self.callback)
        if font is None:
            # cancel should restore the old font
            font = current_font
        self.viewer.antic_font_data = font


class view_add_viewer(ViewerListAction):
    def prune_viewers(self, viewer_cls):
        return viewer_cls.viewer_category != "Emulator"

    def calc_list_items(self):
        subset = [v for v in get_viewers().values() if self.prune_viewers(v)]
        subset.sort(key=lambda v: v.ui_name)
        return subset

    def calc_name(self, action_key):
        viewer = self.get_item(action_key)
        return viewer.ui_name

    def perform(self, action_key):
        viewer = self.get_item(action_key)
        self.editor.add_viewer(viewer)


class view_add_emulation_viewer(view_add_viewer):
    def prune_viewers(self, viewer_cls):
        return viewer_cls.viewer_category == "Emulator"



import os
import time

import wx
import numpy as np

from atrip import Collection, Container, Segment, errors

from sawx.document import SawxDocument
from sawx.utils.nputil import to_numpy
from sawx.events import EventHandler

from .. import emulator as em
from .. import errors
from ..document import DiskImageDocument

import logging
log = logging.getLogger(__name__)


class EmulationTimer(wx.Timer):
    def __init__(self, document):
        self.document = document
        wx.Timer.__init__(self)

    def Notify(self):
        self.document.ready_for_next_frame()


class EmulationDocument(DiskImageDocument):
    """A wrapper around the low-level `Emulator` instance that manages
    the timing when running the emulator.

    Only one instance of a particular emulator type is allowed at any one time.
    The low level C code used in most emulators can't run multiple instances of
    themselves, e.g. atari800 uses global variables for most of its state.
    """

    # Class attributes

    emulation_timer = None

    metadata_extension = ".omniemu"

    # Lookup mapping of emulator ui_name to actual emulator instance to make
    # sure only one emulator of a particular type is used at any one time. I
    # think that's a livable restriction for now.
    emulator_document = {}

    def __init__(self, file_metadata, emulator_type, emulator, source_document=None):
        super().__init__(file_metadata)
        self.source_document = source_document
        self.boot_segment = None
        self.emulator_type = emulator_type
        self.emulator = emulator
        self.skip_frames_on_boot = 0
        self.pause_emulator_on_boot = False
        self.emu_container = None
        self.collection = None

        # Update the graphic screen
        self.priority_level_refresh_event = EventHandler(self)
        self.emulator_breakpoint_event = EventHandler(self)
        self.emulator_update_screen_event = EventHandler(self)
        self.framerate = 1/60.0
        self.tickrate = 1/60.0
        self.last_update_time = 0.0

    #### dunder methods

    def __str__(self):
        return f"EmulationDocument: mime={self.mime}, {self.uri}. {self.emulator_type} source={self.source_document}"

    @classmethod
    def create_document(cls, source_document, emulator_type, skip_frames_on_boot=None, extra_metadata=None):
        try:
            emu_cls = em.find_emulator(emulator_type)
        except errors.UnknownEmulatorError:
            if not emulator_type:
                # if no value specified, try to determine from binary data
                try:
                    emu_cls = em.guess_emulator(source_document)
                except errors.UnknownEmulatorError:
                    emu_cls = em.default_emulator
            else:
                # if emulator name specified but not known, return error
                raise RuntimeError(f"Unknown emulator {emulator_type}")
        if emu_cls is None:
            raise RuntimeError(f"Couldn't determine an emulator for:\n\n{source_document.uri}\n\nand no default emulator specified. Chose an emulator in preferences")
        emu_doc = cls.emulator_document.get(emu_cls.ui_name, None)
        if emu_doc is not None:
            error = errors.EmulatorInUseError(f"Only one {emu_doc.emulator.ui_name} emulator can run at one time.")
            error.current_emulator_document = emu_doc
            raise error
        emu = emu_cls()
        emu.configure_emulator()
        log.debug(f"emulator changed to {emu}")
        doc = cls(None, emulator_type, emu, source_document)
        if skip_frames_on_boot is not None:
            doc.skip_frames_on_boot = skip_frames_on_boot
        elif source_document.skip_frames_on_boot > 0:
            doc.skip_frames_on_boot = source_document.skip_frames_on_boot
        if extra_metadata:
            doc.restore_save_points_from_dict(extra_metadata)
        cls.pause_emulator_on_boot = source_document.pause_emulator_on_boot
        cls.emulator_document[emu_cls.ui_name] = doc
        return doc

    @property
    def timer_delay(self):
        # wxpython delays are in milliseconds
        return self.tickrate * 1000

    @property
    def emulator_running(self):
        return self.emulation_timer is not None and self.emulation_timer.IsRunning()

    @property
    def emulator_paused(self):
        return self.emulation_timer is not None and not self.emulation_timer.IsRunning()

    #### serialization methods

    def restore_extra_from_dict(self, e):
        DiskImageDocument.restore_extra_from_dict(self, e)
        self.restore_save_points_from_dict(e)

    def restore_save_points_from_dict(self, e):
        # emulator type will already be set at document creation time
        if 'restart_tree' in e:
            self.emulator.restart_tree.restore_from_dict(e)
            self.emulator.configure_io_arrays()
            self.create_segments()
        if 'current_restart' in e:
            self.emulator.restore_from_dict(e['current_restart'])
        self.skip_frames_on_boot = e.get('skip_frames_on_boot', 0)

    def serialize_extra_to_dict(self, mdict):
        DiskImageDocument.serialize_extra_to_dict(self, mdict)

        mdict["emulator_type"] = self.emulator.name
        mdict.update(self.emulator.restart_tree.serialize_to_dict())
        mdict["current_restart"] = self.emulator.serialize_to_dict()
        mdict["skip_frames_on_boot"] = self.skip_frames_on_boot

    def save_to_uri(self, uri, editor, saver=None, save_metadata=True):
        # save both the source document and its .omnivore metadata and...
        self.source_document.save_to_uri(uri, editor, saver, save_metadata)

        # the emulator metadata .omniemu from this emulator document
        if save_metadata:
            self.save_metadata_to_uri(uri, editor)

    ##### Initial viewer defaults

    def find_initial_visible_segment(self, linked_base, default=0):
        for segment in self.segments:
            log.debug(f"Looking for Main Memory segment: {segment}")
            if segment.name == "Main Memory":
                linked_base.find_segment(segment, refresh=False)
                break
        else:
            log.warning("Didn't find Main Memory segment in emulator")

    #####

    def boot(self, segment=None):
        emu = self.emulator
        emu.configure_emulator([])
        if segment is None:
            segment = emu.find_default_boot_segment(self.source_document.segments)

        if segment is not None:
            boot_data = segment.data
            origin = segment.origin
        else:
            raise errors.EmulatorError(f"Can't find bootable segment in {self.source_document}")
        emu.boot_from_segment(segment)
        for i in range(self.skip_frames_on_boot):
            emu.next_frame()
        self.create_segments()
        self.create_timer()
        if not self.pause_emulator_on_boot:
            self.start_timer()

    def load(self, segment=None):
        SawxDocument.load(self, segment)

    def create_segments(self):
        emu = self.emulator
        ec = Container(emu.raw_array, force_numpy_data=True)
        for offset, count, origin, name in emu.save_state_memory_blocks:
            s = Segment(ec, offset, origin, name, length=count)
            ec.segments.append(s)
        collection = Collection(emu.ui_name, container=ec, guess=False)
        log.debug(f"Emulator: {emu} collection:{collection.verbose_info}")
        self.load_collection(collection, self.file_metadata)

    #### cleanup

    def halt_background_processing(self):
        print(f"STOPPING TIMERS! {self}")
        self.stop_timer()
        try:
            emu_doc = self.emulator_document.pop(self.emulator.ui_name)
            if emu_doc != self:
                print(f"Weird: emulator not same as remembered!")
        except KeyError:
            log.error(f"Emulator instance not found! Weird.")
        else:
            emu = self.emulator
            print(f"Removing emulator {emu}")
            emu.prepare_destroy()

    def prepare_destroy(self):
        pass

    #### timer control

    def create_timer(self):
        cls = self.__class__
        if cls.emulation_timer is None:
            cls.emulation_timer = EmulationTimer(self)

    def start_timer(self, repeat=False, delay=None, forceupdate=True):
        if not self.emulation_timer.IsRunning():
            self.emulation_timer.StartOnce(self.framerate * 1000)

    def stop_timer(self, repeat=False, delay=None, forceupdate=True):
        self.emulation_timer.Stop()

    def ready_for_next_frame(self):
        now = time.time()
        breakpoint = self.emulator.next_frame()
        frame_number = self.emulator.status['frame_number']
        log.debug(f"showing frame {frame_number}, breakpoint_id={breakpoint}")
        if breakpoint is None:
            self.emulator_update_screen_event(True)
            self.priority_level_refresh_event(True)
            after = time.time()
            delta = after - now
            if delta > self.framerate:
                next_time = self.framerate * .8
            elif delta < self.framerate:
                next_time = self.framerate - delta
            log.debug(f"now={now} show={after-now} delta={delta} framerate={self.framerate} next_time={next_time}")
            if next_time <= 0.001:
                log.warning("need to drop frames!")
                next_time = .001
            self.emulation_timer.StartOnce(next_time * 1000)
        else:
            self.emulator_update_screen_event(True)
            self.priority_level_refresh_event(100)
            print(f"generating breakpoint event: {breakpoint} at cycles={self.emulator.cycles_since_power_on}")
            self.emulator_breakpoint_event(breakpoint)
            self.stop_timer()
        self.last_update_time = now

    #### emulator commands

    def pause_emulator(self):
        print("pause")
        if not self.emulator_paused:
            emu = self.emulator
            self.stop_timer()
            emu.cpu_history_show_next_instruction()
            self.emulator_update_screen_event(True)
            self.priority_level_refresh_event(100)

    def resume_emulator(self):
        print("resume")
        if self.emulator_paused:
            self.emulator.begin_restart()
            self.start_timer()
            self.emulator_update_screen_event(True)

    def debugger_step(self):
        print("stepping")
        self.emulator.step_into(1)
        self.start_timer()

    def debugger_break_vbi_start(self, count=1):
        print("stepping to VBI")
        self.emulator.break_vbi_start(count)
        self.start_timer()

    def debugger_break_vbi_end(self, count=1):
        print("stepping to end of VBI")
        self.emulator.break_vbi_end(count)
        self.start_timer()

    def debugger_break_dli_start(self, count=1):
        print("stepping to DLI")
        self.emulator.break_dli_start(count)
        self.start_timer()

    def debugger_break_dli_end(self, count=1):
        print("stepping to end of DLI")
        self.emulator.break_dli_end(count)
        self.start_timer()

    def debugger_break_next_scan_line(self, count=1):
        print("stepping to next scan line")
        self.emulator.break_scan_line(count)
        self.start_timer()

    def debugger_count_frames(self, number=1):
        print(f"counting {number} frames")
        self.emulator.count_frames(number)
        self.start_timer()

    def history_previous(self):
        emu = self.emulator
        try:
            desired = emu.get_previous_history(emu.current_frame_number)
        except IndexError:
            log.warning("No previous frame")
        else:
            self.checkpoint_restore(emu.current_restart, desired)

    def history_next(self):
        emu = self.emulator
        try:
            desired = emu.get_next_history(emu.current_frame_number)
        except IndexError:
            log.warning(f"No next frame: current = {emu.current_frame_number}")
        else:
            self.checkpoint_restore(emu.current_restart, desired)

    def checkpoint_restore(self, restart_number, frame_number):
        emu = self.emulator
        emu.restore_restart(restart_number, frame_number)
        frame_number = self.emulator.status['frame_number']
        log.debug(f"showing frame {frame_number}")
        self.emulator_update_screen_event(True)
        self.priority_level_refresh_event(100)
